import matplotlib.lines as mlines
import numpy as np

from relsad.load.bus import CostFunction
from relsad.StatDist import StatDist, StatDistType, UniformParameters
from relsad.Time import Time, TimeUnit
from relsad.utils import (
    convert_yearly_fail_rate,
    interpolate,
    random_choice,
    eq,
)

from .Component import Component


class Bus(Component):
    """
    Common base class for all distribution buses

    ...

    Attributes
    ----------
    name : string
        Name of the bus
    n_customers : int
        Number of customers connected to that bus
    coordinate : list
        Coordinate of the bus
    s_ref : float
        Apperent power reference [MVa]
    cost : float
        The energy shedding cost of the bus
    cost_functions : list
        List of power load cost functions
    pload_data : list
        List of active power load data
    qload_data : list
        List of reactive power load data
    ZIP : list
        List showing the ZIP load model
    p_load_downstream : float
        Active accumulated power load at node
    q_load_downstream : float
        Reactive accumulated power load at node
    p_loss_downstream : float
        Active accumulated power line loss at node
    q_loss_downstream : float
        Reactive accumulated power line loss at node
    voang : float
        Voltage angle [rad]
    vomag : float
        Voltage magnitude pu
    pload : float
        The active power load at the bus [MW]
    qload : float
        The reactive power load at the bus [MVar]
    pload_pu : float
        The active power load at the bus in pu
    qload_pu : float
        The reactive power load at the bus in pu
    pprod : float
        The active generated power at the bus [MW]
    qprod : float
        The reactive generated power at the bus [MVar]
    pprod_pu : float
        The active generated power at the bus in pu
    qprod_pu : float
        The reactive generated power at the bus in pu
    is_slack : bool
        Tells if the given bus is a slack bus or not
    toline : Line
        Tells which line that is going into the bus
    fromline : Line
        Tells which line that is going out of the bus
    toline_list : list
        List of lines going into the bus
    fromline_list : list
        List of lines going from the bus
    nextbus : List
        List of neighbor buses
    connected_lines : List
        List of connected lines
    parent_network : PowerNetwork
        Parent network of the bus
    fail_rate_per_year : float
        The failure rate per year for the transformer at the bus
    repair_time_dist : StatDist
        The repair time of the transformer at the bus [hours/fault]
    p_energy_shed_stack : float
        The amount of shedded active energy at the bus
        in the current sequence
    q_energy_shed_stack : float
        The amount of shedded reactive energy at the bus
        in the current sequence
    acc_p_energy_shed : float
        The accumulated amount of shedded active energy at the bus
        for the entire simulation
    acc_q_energy_shed : float
        The accumulated amount of shedded reactive energy at the bus
        for the entire simulation
    acc_outage_time : Time
        The accumulated outage time of the transformer at the bus
        for the entire simulation
    avg_fail_rate : float
        The average failure rate of the transformer at the bus
        for the entire simulation
    avg_outage_time : Time
        The average outage time of the transformer at the bus
        for the entire simulation
    num_consecutive_interruptions : float
        The current number of consecutive interruptions the bus experiences
    interruption_fraction : float
        The current fraction of interruption experienced by the bus
    curr_interruptions : float
        Current number of interruptions experienced by the bus
    acc_interruptions : float
        Accumulated number of interruptions experienced by the bus
    trafo_failed : bool
        Failure status of the transformer
    remaining_outage_time : Time
        The remaining outage time of the bus
    prod : Production
        The Production unit at the bus
    ev_park : EVPark
        The EVPark at the bus
    battery : Battery
        The Battery unit at the bus
    history : dict
        Dictonary that stores the sequential simulation history variables
    monte_carlo_history : dict
        Dictonary that stores the Monte Carlo simulation history variables


    Methods
    ----------
    reset_load_and_production_attributes()
        Resets the load and generation at the bus
    reset_load()
        Resets the load at the bus by setting the load to 0
    reset_prod()
        Resets the generation at the bus by setting the generation to 0
    add_load_data(pload_data, qload_data, cost_function)
        Adds load data to the bus
    prepare_load_data(time_indices)
        Prepares the load data for the current time step configuration
    add_load(pload, qload)
        Adds load to the bus
    set_load_and_cost(inc_idx)
        Sets the bus load and cost in MW based on load and cost profiles in the current increment
    get_load()
        Retuns the current load at the bus in MW
    draw_repair_time(dt)
        Decides and returns the repair time of the trafo based on a statistical distribution
    trafo_fail(dt)
        Sets the transformer status to failed, load and generation at the node are set to zero
    trafo_not_fail()
        Sets the transformer to not failed
    get_battery()
        Returns the battery at the bus
    get_production()
        Returns the generation unit at the bus
    update_fail_status(dt)
        Updates the fail status of the transformer. Sets the fail status to failed if the transformer is failed or the fail status to not failed if the transformer is not failed
    set_slack()
        Sets a bus to slack bus
    print_status()
        Prints the status of the bus
    initialize_history()
        Initializes the history variables
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute at the bus
    set_cost(cost)
        Sets the specificed interruption cost related to Cost of Energy Not Supplied at the bus
    get_cost()
        Returns the specificed interruption cost related to Cost of Energy Not Supplied at the bus
    shed_load(dt)
        Sheds load at the bus and resets the load. The shedded load is added to a stack for the bus
    clear_energy_shed_stack()
        Resets the energy.shed stack for the bus
    add_random_instance(random_gen)
        Adds global random seed
    get_avg_fail_rate(curr_time)
        Returns the average failure rate of the transformer at the bus
    reset_status(save_flag)
        Resets and sets the status of the class parameters
    add_to_energy_shed_stack(p_load, q_load, dt)
        Adds the shedded load to the energy.shed stack at the bus
    reset_load_flow_data()
        Resets the variables used in the load flow analysis
    get_monte_carlo_history(attribute)
        Returns a specified history variable from the Monte Carlo simulation
    get_neighbor_buses()
        Returns the neighboring buses of the bus
    """

    ## Visual attributes
    marker = "|"
    size = 4**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=2,
        markersize=size,
        linestyle="None",
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        n_customers: int = 1,
        coordinate: list = [0, 0],
        ZIP=[0.0, 0.0, 1.0],
        s_ref: float = 1,  # MVA
        is_slack: bool = False,
        fail_rate_per_year: float = 0.0,
        repair_time_dist: StatDist = StatDist(
            stat_dist_type=StatDistType.UNIFORM_FLOAT,
            parameters=UniformParameters(
                min_val=0.0,
                max_val=0.0,
            ),
        ),
    ):

        ## Informative attributes
        self.name = name
        self.n_customers = n_customers
        self.coordinate = coordinate

        ## Power flow attributes
        self.s_ref = s_ref
        self.cost = 0  # cost
        self.cost_functions = []
        self.pload_data = []
        self.qload_data = []
        self.ZIP = ZIP
        self.p_load_downstream = 0.0  # Active accumulated load at node
        self.q_load_downstream = 0.0  # Reactive accumulated load at node
        self.p_loss_downstream = 0.0  # Active accumulated line loss at node
        self.q_loss_downstream = 0.0  # Active accumulated line loss at node
        self.voang = 0.0
        self.vomag = 1.0
        self.pload = 0
        self.qload = 0
        self.pload_pu = 0
        self.qload_pu = 0
        self.pprod = 0
        self.qprod = 0
        self.pprod_pu = 0
        self.qprod_pu = 0

        ## Topological attributes
        self.is_slack = is_slack
        self.toline = None
        self.fromline = None
        self.toline_list = list()
        self.fromline_list = list()
        self.nextbus = list()
        self.connected_lines = list()
        self.parent_network = None

        ## Reliabilility attributes
        self.fail_rate_per_year = fail_rate_per_year  # failures per year
        self.repair_time_dist = repair_time_dist
        self.p_energy_shed_stack = 0
        self.q_energy_shed_stack = 0
        self.acc_p_energy_shed = 0
        self.acc_q_energy_shed = 0
        self.acc_outage_time = Time(0)
        self.avg_fail_rate = 0
        self.avg_outage_time = Time(0)
        self.num_consecutive_interruptions = 0
        self.interruption_fraction = 0
        self.curr_interruptions = 0
        self.acc_interruptions = 0

        ## Status attribute
        self.trafo_failed = False
        self.remaining_outage_time = Time(0)

        ## Production, EV park and battery
        self.prod = None
        self.ev_park = None
        self.battery = None

        ## History
        self.history = {}
        self.monte_carlo_history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Bus(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, Bus)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def reset_load_and_prod_attributes(self):
        """
        Resets the load and generation at the bus

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.reset_load()
        self.reset_prod()

    def reset_load(self):
        """
        Resets the load at the bus by setting the load to 0

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.pload = 0
        self.qload = 0
        self.pload_pu = 0
        self.qload_pu = 0

    def reset_prod(self):
        """
        Resets the generation at the bus by setting the generation to 0

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.pprod = 0
        self.qprod = 0
        self.pprod_pu = 0
        self.qprod_pu = 0

    def add_load_data(
        self,
        pload_data: np.ndarray,
        qload_data: np.ndarray = None,
        cost_function: CostFunction = CostFunction(A=1, B=0),
    ):
        """
        Adds load data to the bus

        Parameters
        ----------
        pload_data : np.ndarray
            Active power load array
        qload_data : np.ndarray
            Reactive power load array
        cost_function : CostFunction
            Load cost function

        Returns
        ----------
        None

        """
        self.cost_functions.append(cost_function)
        self.pload_data.append(pload_data)
        if qload_data is None:
            self.qload_data.append(np.zeros_like(pload_data))
        else:
            self.qload_data.append(qload_data)

    def prepare_load_data(
        self,
        time_indices: np.ndarray,
    ):
        """
        Prepares the load data for the current time step configuration

        Parameters
        ----------
        time_indices : np.ndarray
            Time indices used to discretize the load data

        Returns
        ----------
        None

        """
        for i, pload_array in enumerate(self.pload_data):
            self.pload_data[i] = interpolate(
                array=pload_array,
                time_indices=time_indices,
            )
        for i, qload_array in enumerate(self.qload_data):
            self.qload_data[i] = interpolate(
                array=qload_array,
                time_indices=time_indices,
            )

    def add_load(
        self,
        pload: float,
        qload: float,
    ):
        """
        Adds load to the bus

        Parameters
        ----------
        pload : float
            Active power
        qload : float
            Reactive power

        Returns
        ----------
        None

        """
        # MW and MVar
        self.pload += pload
        self.qload += qload
        # Per unit
        self.pload_pu = self.pload / self.s_ref
        self.qload_pu = self.qload / self.s_ref

    def set_load_and_cost(self, inc_idx: int):
        """
        Sets the bus load and cost in MW based on load and cost profiles
        in the current increment

        Parameters
        ----------
        inc_idx : int
            Index of the current increment

        Returns
        ----------
        None

        """
        self.reset_load()
        default_cost = 1e8
        type_cost = 0
        for i, cost_function in enumerate(self.cost_functions):
            if (
                not self.pload_data[i] is None
                and not self.qload_data[i] is None
            ):
                type_cost = max(
                    type_cost,
                    cost_function.A + cost_function.B * 1,
                )
                self.add_load(
                    pload=self.pload_data[i][inc_idx] * self.n_customers,
                    qload=self.qload_data[i][inc_idx] * self.n_customers,
                )
        if type_cost > 0:
            self.set_cost(type_cost)
        else:
            self.set_cost(default_cost)

    def get_load(self):
        """
        Returns the current load at the bus in MW

        Parameters
        ----------
        None

        Returns
        ----------
        pload : float
            The active load at the bus
        qload : float
            The reactive load at the bus

        """
        return self.pload, self.qload

    def draw_repair_time(self, dt: Time):
        """
        Decides and returns the repair time of the trafo based on a statistical distribution

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        return Time(
            self.repair_time_dist.draw(
                random_instance=self.ps_random,
                size=1,
            )[0],
            dt.unit,
        )

    def trafo_fail(self, dt: Time):
        """
        Sets the transformer status to failed, load and generation at the bus are set to zero

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.trafo_failed = True
        self.remaining_outage_time = self.draw_repair_time(dt)
        self.shed_load(dt)
        if self.prod is not None:
            self.prod.reset_prod()

    def trafo_not_fail(self):
        """
        Sets the transformer status to not failed

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.trafo_failed = False

    def get_battery(self):
        """
        Returns the battery at the bus

        Parameters
        ----------
        None

        Returns
        ----------
        battery : Battery
            Returns the battery at the bus, none if there is no battery at the bus

        """
        return self.battery

    def get_production(self):
        """
        Returns the generation unit at the bus

        Parameters
        ----------
        None

        Returns
        ----------
        prod : Production
            Returns the generation at the bus, none if there is no battery at the bus

        """
        return self.prod

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status of the transformer.
        Sets the fail status to failed if the transformer is failed
        or the fail status to not failed if the transformer is not failed

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.trafo_failed:
            self.remaining_outage_time -= dt
            if self.remaining_outage_time <= Time(0):
                self.trafo_not_fail()
                self.remaining_outage_time = Time(0)
            else:
                self.shed_load(dt)
                if self.prod is not None:
                    self.prod.reset_prod()
        else:
            p_fail = convert_yearly_fail_rate(self.fail_rate_per_year, dt)
            if random_choice(self.ps_random, p_fail):
                self.trafo_fail(dt)
            else:
                self.trafo_not_fail()

    def set_slack(self):
        """
        Sets a bus to slack bus

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_slack = True

    def print_status(self):
        """
        Prints the status of the bus

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        print(
            "name: {:3s}, trafo_failed={}, pload={:.4f}, "
            "is_slack={}".format(
                self.name, self.trafo_failed, self.pload, self.is_slack
            )
        )

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["pload"] = {}
        self.history["qload"] = {}
        self.history["pprod"] = {}
        self.history["qprod"] = {}
        self.history["remaining_outage_time"] = {}
        self.history["trafo_failed"] = {}
        self.history["p_energy_shed_stack"] = {}
        self.history["acc_p_energy_shed"] = {}
        self.history["q_energy_shed_stack"] = {}
        self.history["acc_q_energy_shed"] = {}
        self.history["voang"] = {}
        self.history["vomag"] = {}
        self.history["avg_fail_rate"] = {}
        self.history["avg_outage_time"] = {}
        self.history["acc_outage_time"] = {}
        self.history["interruption_fraction"] = {}
        self.history["acc_interruptions"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        # Update accumulated energy.shed for bus
        self.acc_p_energy_shed += self.p_energy_shed_stack
        self.acc_q_energy_shed += self.q_energy_shed_stack
        dt = curr_time - prev_time if prev_time is not None else curr_time
        self.acc_outage_time += dt if self.p_energy_shed_stack > 0 else Time(0)
        self.avg_outage_time = (
            Time(
                self.acc_outage_time / curr_time, curr_time.unit
            )
            if curr_time != Time(0, curr_time.unit)
            else Time(0, curr_time.unit)
        )
        self.avg_fail_rate = self.get_avg_fail_rate(curr_time)
        # Accumulate fraction of interupted customers
        self.interruption_fraction = (
            self.p_energy_shed_stack / (self.pload * dt.get_hours())
            if not eq(self.pload, 0) and dt.get_hours() > 0
            else 0
        )

        if self.interruption_fraction > 0:
            self.curr_interruptions += self.interruption_fraction
            self.num_consecutive_interruptions += 1
        else:
            if self.num_consecutive_interruptions >= 1:
                self.acc_interruptions += (
                    self.curr_interruptions
                    / self.num_consecutive_interruptions
                )
            self.curr_interruptions = 0
            self.num_consecutive_interruptions = 0

        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["pload"][time] = self.pload
            self.history["qload"][time] = self.qload
            self.history["pprod"][time] = self.pprod
            self.history["qprod"][time] = self.qprod
            self.history["remaining_outage_time"][
                time
            ] = self.remaining_outage_time.get_unit_quantity(curr_time.unit)
            self.history["trafo_failed"][time] = self.trafo_failed
            self.history["p_energy_shed_stack"][
                time
            ] = self.p_energy_shed_stack
            self.history["acc_p_energy_shed"][
                time
            ] = self.acc_p_energy_shed
            self.history["q_energy_shed_stack"][
                time
            ] = self.q_energy_shed_stack
            self.history["acc_q_energy_shed"][
                time
            ] = self.acc_q_energy_shed
            self.history["voang"][time] = self.voang
            self.history["vomag"][time] = self.vomag
            self.history["avg_fail_rate"][
                time
            ] = self.avg_fail_rate  # Average failure rate (lamda_s)
            self.history["avg_outage_time"][
                time
            ] = self.avg_outage_time.get_unit_quantity(
                curr_time.unit
            )  # Average outage time (r_s)
            self.history["acc_outage_time"][
                time
            ] = self.acc_outage_time.get_unit_quantity(
                curr_time.unit
            )  # Accumulated outage time
            self.history["interruption_fraction"][
                time
            ] = self.interruption_fraction
            self.history["acc_interruptions"][
                time
            ] = self.acc_interruptions
        self.clear_energy_shed_stack()

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute at the bus

        Parameters
        ----------
        attribute : str
            Bus attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute
        """
        return self.history[attribute]

    def set_cost(self, cost: float):
        """
        Sets the specificed interruption cost related to Cost of Energy Not Supplied at the bus

        Parameters
        ----------
        cost : float
            The specificed interruption cost

        Returns
        ----------
        None

        """
        self.cost = cost

    def get_cost(self):
        """
        Returns the specificed interruption cost related to Cost of Energy Not Supplied at the bus

        Parameters
        ----------
        None

        Returns
        ----------
        cost : float
            The specificed interruption cost

        """
        return self.cost

    def shed_load(self, dt: Time):
        """
        Sheds load at the bus and resets the load. The shedded load is added to a stack for the bus

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.add_to_energy_shed_stack(
            self.pload,  # MW
            self.qload,  # MW
            dt,
        )
        self.reset_load()

    def clear_energy_shed_stack(self):
        """
        Resets the energy.shed stack for the bus

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.p_energy_shed_stack = 0
        self.q_energy_shed_stack = 0

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def get_avg_fail_rate(self, curr_time: Time):
        """
        Returns the average failure rate of the transformer at the bus

        Parameters
        ----------
        None

        Returns
        ----------
        avg_fail_rate : float
            The average failure rate of the transformer at the bus

        """
        fail_rate = self.fail_rate_per_year
        if self.parent_network is not None:
            for line in self.parent_network.get_lines():
                fail_rate += line.fail_rate_per_year
        avg_fail_rate = (
            fail_rate / curr_time.get_years()
            if curr_time.get_years() > 0
            else 0
        )
        return avg_fail_rate

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.trafo_failed = False
        self.remaining_outage_time = Time(0)
        self.acc_outage_time = Time(0)
        self.reset_load_and_prod_attributes()
        self.cost = 0  # cost
        self.clear_energy_shed_stack()
        # Accumulated energy.shed for bus
        self.acc_p_energy_shed = 0
        self.acc_q_energy_shed = 0
        self.num_consecutive_interruptions = 0
        self.interruption_fraction = 0
        self.curr_interruptions = 0
        self.acc_interruptions = 0
        if save_flag:
            self.initialize_history()

    def add_to_energy_shed_stack(self, p_load: float, q_load: float, dt: Time):
        """
        Adds the shedded load to the energy.shed stack at the bus

        Parameters
        ----------
        p_load : float
            The active power at the bus
        q_load : float
            The reactive power at the bus
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.p_energy_shed_stack += p_load * dt.get_hours()  # MWh
        self.q_energy_shed_stack += q_load * dt.get_hours()  # MWh

    def reset_load_flow_data(self):
        """
        Resets the variables used in the load flow analysis

        Parameters
        ---------
        None

        Returns
        --------
        None
        """
        self.p_load_downstream = 0.0  # Active accumulated load at node
        self.q_load_downstream = 0.0  # Reactive accumulated load at node
        self.p_loss_downstream = 0.0  # Active accumulated line loss at node
        self.q_loss_downstream = 0.0  # Active accumulated line loss at node
        self.voang = 0.0
        self.vomag = 1.0

    def get_monte_carlo_history(self, attribute):
        """
        Returns a specified history variable from the Monte Carlo simulation

        Parameters
        ---------
        attribute : str
            Bus attribute

        Returns
        --------
        monte_carlo_history[attribute] : str
            The specified history variable from the Monte Carlo simulation
        """
        return self.monte_carlo_history[attribute]

    def get_neighbor_buses(self):
        """
        Returns the neighboring buses of the bus

        Parameters
        ---------
        None

        Returns
        --------
        neighbor_buses : list
            List of neighboring buses
        """
        neighbor_buses = []
        for line in self.connected_lines:
            if line.connected is True:
                for bus in [line.fbus, line.tbus]:
                    if bus not in neighbor_buses and bus != self:
                        neighbor_buses.append(bus)

        return neighbor_buses



import os

import pandas as pd


def save_monte_carlo_history_from_dict(
    save_dict: dict,
    comp_list: list,
    attribute: str,
    save_dir: str,
):
    """
    Saves history from dictionary to files

    Parameters
    ----------
    save_dict : dict
        Dictionary with simulation results
    comp_list : list
        A list of system componets
    attribute : str
        A system attribute
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """

    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    df = pd.DataFrame()
    for comp in comp_list:
        data = save_dict[comp.name][attribute]
        df[comp] = data.values()
    df.index = data.keys()
    df.index.name = "it"
    df.to_csv(os.path.join(save_dir, attribute + ".csv"))



import os

import pandas as pd

from relsad.Time import TimeUnit


def save_history(
    obj_list: list,
    attribute: str,
    time_unit: TimeUnit,
    save_dir: str,
):
    """
    Saves history to files

    Parameters
    ----------
    obj_list : list
        A list of relsad objects
    attribute : str
        A system attribute
    time_unit : TimeUnit
        The time unit of the simulation
    save_dir : str
        The saving directory

    Returns
    ----------
    None

    """
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    df = pd.DataFrame()
    for obj in obj_list:
        data = obj.get_history(attribute)
        df[obj] = data.values()
    df.index = data.keys()
    df.index.name = time_unit.name
    df.to_csv(os.path.join(save_dir, attribute + ".csv"))



from enum import Enum
from numbers import Number


class TimeUnit(Enum):
    """
    Time unit

    Attributes
    ----------
    SECOND : int
        Second
    MINUTE : int
        Minute
    HOUR : int
        Hour
    DAY : int
        Day
    WEEK : int
        Week
    MONTH : int
        Month
    YEAR : int
        Year
    """

    SECOND = 1
    MINUTE = 2
    HOUR = 3
    DAY = 4
    WEEK = 5
    MONTH = 6
    YEAR = 7


class Time:

    """
    Time utility class

    ...

    Attributes
    ----------
    SEC_per_MIN : float
        Seconds in a minute
    MIN_per_HOUR : float
        Minutes in an hour
    HOUR_per_DAY : float
        Hours in a day
    DAY_per_WEEK : float
        Days in a week
    WEEK_per_MONTH : float
        Weeks in a month
    MONTH_per_YEAR : float
        Months in a year
    quantity : float
        Time quantity, the quantity of a time unit
    unit : TimeUnit
        The time unit

    Methods
    ----------
    convert_unit(unit)
        Converts the time based on the sat time unit to the correct time
    get_unit_quantity(unit)
        Returns the time quantity of the class instance in the given unit
    get_seconds()
        Returns the time quantity in seconds
    get_minutes()
        Returns the time quantity in minutes
    get_hours()
        Returns the time quantity in hours
    get_days()
        Returns the time quantity in days
    get_months()
        Returns the time quantity in months
    get_years()
        Returns the time quantity in year

    """

    SEC_per_MIN = 60
    MIN_per_HOUR = 60
    HOUR_per_DAY = 24
    DAY_per_WEEK = 7
    WEEK_per_MONTH = 4.3452
    MONTH_per_YEAR = 12

    def __init__(self, quantity: float, unit: TimeUnit = TimeUnit.HOUR):
        if not isinstance(quantity, Number):
            raise Exception("The time quantity must be a number.")
        self.quantity = quantity
        self.unit = unit

    def __str__(self):
        if self.unit == TimeUnit.SECOND:
            return f"SEC_{self.quantity}"
        elif self.unit == TimeUnit.MINUTE:
            return f"MIN_{self.quantity}"
        elif self.unit == TimeUnit.HOUR:
            return f"HOUR_{self.quantity}"
        elif self.unit == TimeUnit.DAY:
            return f"DAY_{self.quantity}"
        elif self.unit == TimeUnit.WEEK:
            return f"WEEK_{self.quantity}"
        elif self.unit == TimeUnit.MONTH:
            return f"MONTH_{self.quantity}"
        elif self.unit == TimeUnit.YEAR:
            return f"YEAR_{self.quantity}"

    def __repr__(self):
        if self.unit == TimeUnit.SECOND:
            return f"SEC_{self.quantity}"
        elif self.unit == TimeUnit.MINUTE:
            return f"MIN_{self.quantity}"
        elif self.unit == TimeUnit.HOUR:
            return f"HOUR_{self.quantity}"
        elif self.unit == TimeUnit.DAY:
            return f"DAY_{self.quantity}"
        elif self.unit == TimeUnit.WEEK:
            return f"WEEK_{self.quantity}"
        elif self.unit == TimeUnit.MONTH:
            return f"MONTH_{self.quantity}"
        elif self.unit == TimeUnit.YEAR:
            return f"YEAR_{self.quantity}"

    def __hash__(self):
        if self.unit == TimeUnit.SECOND:
            return hash(f"SEC_{self.quantity}")
        elif self.unit == TimeUnit.MINUTE:
            return hash(f"MIN_{self.quantity}")
        elif self.unit == TimeUnit.HOUR:
            return hash(f"HOUR_{self.quantity}")
        elif self.unit == TimeUnit.DAY:
            return hash(f"DAY_{self.quantity}")
        elif self.unit == TimeUnit.WEEK:
            return hash(f"WEEK_{self.quantity}")
        elif self.unit == TimeUnit.MONTH:
            return hash(f"MONTH_{self.quantity}")
        elif self.unit == TimeUnit.YEAR:
            return hash(f"YEAR_{self.quantity}")

    def convert_unit(self, unit: TimeUnit):
        """
        Converts the time based on the sat time unit to the correct time

        Parameters
        ----------
        unit : TimeUnit
            The time unit

        Returns
        ----------
        None

        """
        if unit == TimeUnit.SECOND:
            self.quantity = self.get_seconds()
            self.unit = TimeUnit.SECOND
        elif unit == TimeUnit.MINUTE:
            self.quantity = self.get_minutes()
            self.unit = TimeUnit.MINUTE
        elif unit == TimeUnit.HOUR:
            self.quantity = self.get_hours()
            self.unit = TimeUnit.HOUR
        elif unit == TimeUnit.DAY:
            self.quantity = self.get_days()
            self.unit = TimeUnit.DAY
        elif unit == TimeUnit.WEEK:
            self.quantity = self.get_weeks()
            self.unit = TimeUnit.WEEK
        elif unit == TimeUnit.MONTH:
            self.quantity = self.get_months()
            self.unit = TimeUnit.MONTH
        elif unit == TimeUnit.YEAR:
            self.quantity = self.get_years()
            self.unit = TimeUnit.YEAR

    def get_unit_quantity(self, unit: TimeUnit):
        """
        Returns the time quantity of the class instance in the given unit

        Parameters
        ----------
        unit : TimeUnit
            The time unit

        Returns
        ----------
        time_quantity : float
            The time quantity of the class instance in the given unit

        """
        time_quantity = 0
        if unit == TimeUnit.SECOND:
            time_quantity = self.get_seconds()
        elif unit == TimeUnit.MINUTE:
            time_quantity = self.get_minutes()
        elif unit == TimeUnit.HOUR:
            time_quantity = self.get_hours()
        elif unit == TimeUnit.DAY:
            time_quantity = self.get_days()
        elif unit == TimeUnit.WEEK:
            time_quantity = self.get_weeks()
        elif unit == TimeUnit.MONTH:
            time_quantity = self.get_months()
        elif unit == TimeUnit.YEAR:
            time_quantity = self.get_years()
        return time_quantity

    def __lt__(self, other):
        return isinstance(
            other, self.__class__
        ) and self.quantity < other.get_unit_quantity(self.unit)

    def __le__(self, other):
        return isinstance(
            other, self.__class__
        ) and self.quantity <= other.get_unit_quantity(self.unit)

    def __gt__(self, other):
        return isinstance(
            other, self.__class__
        ) and self.quantity > other.get_unit_quantity(self.unit)

    def __ge__(self, other):
        return isinstance(
            other, self.__class__
        ) and self.quantity >= other.get_unit_quantity(self.unit)

    def __eq__(self, other):
        return isinstance(
            other, self.__class__
        ) and self.quantity == other.get_unit_quantity(self.unit)

    def __ne__(self, other):
        return isinstance(
            other, self.__class__
        ) and self.quantity != other.get_unit_quantity(self.unit)

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise Exception("Wrong type")
        return Time(
            self.quantity + other.get_unit_quantity(self.unit), self.unit
        )

    def __sub__(self, other):
        if not isinstance(other, self.__class__):
            raise Exception("Wrong type")
        return Time(
            self.quantity - other.get_unit_quantity(self.unit), self.unit
        )

    def __truediv__(self, other):
        if not isinstance(other, self.__class__):
            raise Exception("Wrong type")
        if other.get_hours() == 0:
            raise Exception("Other time is zero")
        return self.get_hours() / other.get_hours()

    def get_seconds(self):
        """
        Returns the time quantity in seconds

        Parameters
        ----------
        None

        Returns
        ----------
        seconds : float
            The time quantity in seconds

        """
        seconds = 0
        if self.unit == TimeUnit.SECOND:
            seconds = self.quantity
        elif self.unit == TimeUnit.MINUTE:
            seconds = self.quantity * self.SEC_per_MIN
        elif self.unit == TimeUnit.HOUR:
            seconds = self.quantity * self.SEC_per_MIN * self.MIN_per_HOUR
        elif self.unit == TimeUnit.DAY:
            seconds = (
                self.quantity
                * self.SEC_per_MIN
                * self.MIN_per_HOUR
                * self.HOUR_per_DAY
            )
        elif self.unit == TimeUnit.WEEK:
            seconds = (
                self.quantity
                * self.SEC_per_MIN
                * self.MIN_per_HOUR
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
            )
        elif self.unit == TimeUnit.MONTH:
            seconds = (
                self.quantity
                * self.SEC_per_MIN
                * self.MIN_per_HOUR
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
                * self.WEEK_per_MONTH
            )
        elif self.unit == TimeUnit.YEAR:
            seconds = (
                self.quantity
                * self.SEC_per_MIN
                * self.MIN_per_HOUR
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
                * self.WEEK_per_MONTH
                * self.MONTH_per_YEAR
            )
        return seconds

    def get_minutes(self):
        """
        Returns the time quantity in minutes

        Parameters
        ----------
        None

        Returns
        ----------
        minutes : float
            The time quantity in minutes

        """
        minutes = 0
        if self.unit == TimeUnit.SECOND:
            minutes = self.quantity / self.SEC_per_MIN
        elif self.unit == TimeUnit.MINUTE:
            minutes = self.quantity
        elif self.unit == TimeUnit.HOUR:
            minutes = self.quantity * self.MIN_per_HOUR
        elif self.unit == TimeUnit.DAY:
            minutes = self.quantity * self.MIN_per_HOUR * self.HOUR_per_DAY
        elif self.unit == TimeUnit.WEEK:
            minutes = (
                self.quantity
                * self.MIN_per_HOUR
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
            )
        elif self.unit == TimeUnit.MONTH:
            minutes = (
                self.quantity
                * self.MIN_per_HOUR
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
                * self.WEEK_per_MONTH
            )
        elif self.unit == TimeUnit.YEAR:
            minutes = (
                self.quantity
                * self.MIN_per_HOUR
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
                * self.WEEK_per_MONTH
                * self.MONTH_per_YEAR
            )
        return minutes

    def get_hours(self):
        """
        Returns the time quantity in hours

        Parameters
        ----------
        None

        Returns
        ----------
        hours : float
            The time quantity in hours

        """
        hours = 0
        if self.unit == TimeUnit.SECOND:
            hours = self.quantity / self.SEC_per_MIN / self.MIN_per_HOUR
        elif self.unit == TimeUnit.MINUTE:
            hours = self.quantity / self.MIN_per_HOUR
        elif self.unit == TimeUnit.HOUR:
            hours = self.quantity
        elif self.unit == TimeUnit.DAY:
            hours = self.quantity * self.HOUR_per_DAY
        elif self.unit == TimeUnit.WEEK:
            hours = self.quantity * self.HOUR_per_DAY * self.DAY_per_WEEK
        elif self.unit == TimeUnit.MONTH:
            hours = (
                self.quantity
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
                * self.WEEK_per_MONTH
            )
        elif self.unit == TimeUnit.YEAR:
            hours = (
                self.quantity
                * self.HOUR_per_DAY
                * self.DAY_per_WEEK
                * self.WEEK_per_MONTH
                * self.MONTH_per_YEAR
            )
        return hours

    def get_days(self):
        """
        Returns the time quantity in days

        Parameters
        ----------
        None

        Returns
        ----------
        days : float
            The time quantity in days

        """
        days = 0
        if self.unit == TimeUnit.SECOND:
            days = (
                self.quantity
                / self.SEC_per_MIN
                / self.MIN_per_HOUR
                / self.HOUR_per_DAY
            )
        elif self.unit == TimeUnit.MINUTE:
            days = self.quantity / self.MIN_per_HOUR / self.HOUR_per_DAY
        elif self.unit == TimeUnit.HOUR:
            days = self.quantity / self.HOUR_per_DAY
        elif self.unit == TimeUnit.DAY:
            days = self.quantity
        elif self.unit == TimeUnit.WEEK:
            days = self.quantity * self.DAY_per_WEEK
        elif self.unit == TimeUnit.MONTH:
            days = self.quantity * self.DAY_per_WEEK * self.WEEK_per_MONTH
        elif self.unit == TimeUnit.YEAR:
            days = (
                self.quantity
                * self.DAY_per_WEEK
                * self.WEEK_per_MONTH
                * self.MONTH_per_YEAR
            )
        return days

    def get_weeks(self):
        """
        Returns the time quantity in weeks

        Parameters
        ----------
        None

        Returns
        ----------
        weeks : float
            The time quantity in weeks

        """
        weeks = 0
        if self.unit == TimeUnit.SECOND:
            weeks = (
                self.quantity
                / self.SEC_per_MIN
                / self.MIN_per_HOUR
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
            )
        elif self.unit == TimeUnit.MINUTE:
            weeks = (
                self.quantity
                / self.MIN_per_HOUR
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
            )
        elif self.unit == TimeUnit.HOUR:
            weeks = self.quantity / self.HOUR_per_DAY / self.DAY_per_WEEK
        elif self.unit == TimeUnit.DAY:
            weeks = self.quantity / self.DAY_per_WEEK
        elif self.unit == TimeUnit.WEEK:
            weeks = self.quantity
        elif self.unit == TimeUnit.MONTH:
            weeks = self.quantity * self.WEEK_per_MONTH
        elif self.unit == TimeUnit.YEAR:
            weeks = self.quantity * self.WEEK_per_MONTH * self.MONTH_per_YEAR
        return weeks

    def get_months(self):
        """
        Returns the time quantity in months

        Parameters
        ----------
        None

        Returns
        ----------
        months : float
            The time quantity in months

        """
        months = 0
        if self.unit == TimeUnit.SECOND:
            months = (
                self.quantity
                / self.SEC_per_MIN
                / self.MIN_per_HOUR
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
                / self.WEEK_per_MONTH
            )
        elif self.unit == TimeUnit.MINUTE:
            months = (
                self.quantity
                / self.MIN_per_HOUR
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
                / self.WEEK_per_MONTH
            )
        elif self.unit == TimeUnit.HOUR:
            months = (
                self.quantity
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
                / self.WEEK_per_MONTH
            )
        elif self.unit == TimeUnit.DAY:
            months = self.quantity / self.DAY_per_WEEK / self.WEEK_per_MONTH
        elif self.unit == TimeUnit.WEEK:
            months = self.quantity / self.WEEK_per_MONTH
        elif self.unit == TimeUnit.MONTH:
            months = self.quantity
        elif self.unit == TimeUnit.YEAR:
            months = self.quantity * self.MONTH_per_YEAR
        return months

    def get_years(self):
        """
        Returns the time quantity in years

        Parameters
        ----------
        None

        Returns
        ----------
        years : float
            The time quantity in years

        """
        years = 0
        if self.unit == TimeUnit.SECOND:
            years = (
                self.quantity
                / self.SEC_per_MIN
                / self.MIN_per_HOUR
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
                / self.WEEK_per_MONTH
                / self.MONTH_per_YEAR
            )
        elif self.unit == TimeUnit.MINUTE:
            years = (
                self.quantity
                / self.MIN_per_HOUR
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
                / self.WEEK_per_MONTH
                / self.MONTH_per_YEAR
            )
        elif self.unit == TimeUnit.HOUR:
            years = (
                self.quantity
                / self.HOUR_per_DAY
                / self.DAY_per_WEEK
                / self.WEEK_per_MONTH
                / self.MONTH_per_YEAR
            )
        elif self.unit == TimeUnit.DAY:
            years = (
                self.quantity
                / self.DAY_per_WEEK
                / self.WEEK_per_MONTH
                / self.MONTH_per_YEAR
            )
        elif self.unit == TimeUnit.WEEK:
            years = self.quantity / self.WEEK_per_MONTH / self.MONTH_per_YEAR
        elif self.unit == TimeUnit.MONTH:
            years = self.quantity / self.MONTH_per_YEAR
        elif self.unit == TimeUnit.YEAR:
            years = self.quantity
        return years


class TimeStamp:
    """
    Time stamp utility class

    ...

    Attributes
    ----------
    year : int
        Number of years
    month : int
        Number of months
    day : int
        Number of days
    hour : int
        Number of hours
    minute : int
        Number of minutes
    second : int
        Number of seconds

    Methods
    ----------
    get_hour_of_day(time_passed)
        Returns the hour of day based on the time passed

    """

    def __init__(
        self,
        year: int = 0,
        month: int = 0,
        day: int = 0,
        hour: int = 0,
        minute: int = 0,
        second: int = 0,
    ):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        # Calculate time quantity
        self.time_quantity = Time(0, TimeUnit.SECOND)
        self.time_quantity += Time(year, TimeUnit.YEAR)
        self.time_quantity += Time(month, TimeUnit.MONTH)
        self.time_quantity += Time(day, TimeUnit.DAY)
        self.time_quantity += Time(hour, TimeUnit.HOUR)
        self.time_quantity += Time(minute, TimeUnit.MINUTE)
        self.time_quantity += Time(second, TimeUnit.SECOND)

    def __str__(self):
        return (
            f"TIMESTAMP_Y{self.year}-"
            + f"M{self.month}-"
            + f"D{self.day}_"
            + f"H{self.hour}:"
            + f"M{self.month}:"
            + f"S{self.second}"
        )

    def __repr__(self):
        return (
            f"TIMESTAMP_Y{self.year}-"
            + f"M{self.month}-"
            + f"D{self.day}_"
            + f"H{self.hour}:"
            + f"M{self.minute}:"
            + f"S{self.second}"
        )

    def __hash__(self):
        return hash(
            f"TIMESTAMP_Y{self.year}-"
            + f"M{self.month}-"
            + f"D{self.day}_"
            + f"H{self.hour}:"
            + f"M{self.minute}:"
            + f"S{self.second}"
        )

    def get_hour_of_day(self, time_passed: Time):
        """
        Returns the hour of day based on the time passed

        Parameters
        ----------
        time_passed : Time
            Time passed

        Returns
        ----------
        hour_of_day : int
            The hour of day based on the time passed

        """
        if not isinstance(time_passed, Time):
            raise Exception("Wrong type")
        dup = Time(time_passed.quantity, time_passed.unit)
        if dup.get_years() > 1:
            y = int(dup.get_years())
            dup -= Time(y, TimeUnit.YEAR)
        if dup.get_months() > 1:
            m = int(dup.get_months())
            dup -= Time(m, TimeUnit.MONTH)
        if dup.get_days() > 1:
            d = int(dup.get_days())
            dup -= Time(d, TimeUnit.DAY)
        dup -= Time(self.hour, TimeUnit.HOUR)
        dup += Time(self.minute, TimeUnit.MINUTE)
        dup += Time(self.second, TimeUnit.SECOND)
        hour_of_day = int(dup.get_hours())
        return hour_of_day

    def __sub__(self, other):
        if not isinstance(other, self.__class__):
            raise Exception("Wrong type")
        return (
            Time(self.year - other.year, TimeUnit.YEAR)
            + Time(self.month - other.month, TimeUnit.MONTH)
            + Time(self.day - other.day, TimeUnit.DAY)
            + Time(self.hour - other.hour, TimeUnit.HOUR)
            + Time(self.minute - other.minute, TimeUnit.MINUTE)
            + Time(self.second - other.second, TimeUnit.SECOND)
        )


if __name__ == "__main__":
    t1 = Time(5, TimeUnit.MINUTE)
    print(t1.get_years())
    print(t1.get_hours())
    t2 = Time(1, TimeUnit.HOUR)
    print(t2.get_years())
    print(t2.get_hours())



from relsad.network.components import (
    Bus,
    Line,
    MicrogridController,
    MicrogridMode,
)
from relsad.utils import unique

from .Distribution import Distribution
from .PowerNetwork import PowerNetwork


class Microgrid(PowerNetwork):
    """
    Class defining a microgrid network type

    ...

    Attributes
    ----------
    name : str
        Name of the microgrid
    mode : str
        Which mode the microgrid follows
    buses : list
        List with the buses in the microgrid
    ev_parks : list
        List containing the EV parks in the microgrid
    batteries : list
        List containing the batteries in the microgrid
    productions : list
        List containing the generation units in the microgrid
    lines : list
        List with the lines in the microgrid
    sensors : list
        List containing the sensors in the microgrid
    circuitbreaker : list
        List containing the circuit breakers in the microgrid
    disconnectors : list
        List containing the disconnectors in the microgrid
    intelligent_switches : list
        List containing the intelligent switches in the microgrid
    controller : MicrogridController
        The controller for the microgrid
    comp_list : list
        List containing the components in the microgrid
    comp_dict : dict
        Dictionary containing the components in the microgrid
    distribution_network : PowerNetwork
        The distribution network the microgrid is connected to
    failed_line : Bool
        Boolean value stating whether or not the network includes a failed line
    connected_line : Line
        Line that connects the microgrid to the distribution network
    circuitbreaker : CircuitBreaker
        The circuitbreaker connected to the line
    p_energy_shed : float
        Shedded active power in the microgrid
    acc_p_energy_shed : float
        The accumulated shedded active power in the microgrid
    q_energy_shed : float
        Shedded reactive power in the microgrid
    acc_q_energy_shed : float
        The accumulated shedded reactive power in the microgrid
    sections : Section
        The sections in the microgrid
    history : dict
        Dictionary containing the history variables of the network
    monte_carlo_history : dict
        Dictionary containing the history variables from the monte carlo simulation

    Methods
    ----------
    add_connected_line(connected_line, mode)
        Sets the line connecting the microgrid to overlying network
    add_bus(bus)
        Adding a bus including elements on the bus (battery, generation unit, EV park) to the microgrid
    add_buses(buses)
        Adding buses to the microgrid
    add_line(line)
        Adding a line including elements on the line (sensor, circuit breaker, disconnector) to the microgrid
    add_lines(lines)
        Adding lines to the microgrid
    get_lines()
        Returns the lines in the microgrid
    connect(self)
        Connects the microgrid to the parent distribution network by closing the disconnectors on the microgrid lines
    disconnect()
         Disconnects the microgrid from the parent distribution network by opening the disconnectors on the microgrid lines
    reset_slack_bus()
        Resets the slack bus attribute of the buses in the microgrid
    get_max_load()
        Get the maximum load of the microgrid form the entire load history and returns the max load
    get_monte_carlo_history(attribute)
        Returns the specified history variable from the Monte Carlo simulation
    get_history(attribute)
        Returns the specified history variable
    get_system_load()
        Returns the system load in the microgrid at the current time in MW and MVar
    reset_energy_shed_variables()
        Resets the energy.shed variables

    """

    ## Visual attributes
    color = "orange"

    ## Counter
    counter = 0

    def __init__(
        self,
        distribution_network: Distribution,
        connected_line: Line,
        mode: MicrogridMode = MicrogridMode.SURVIVAL,
    ):
        Microgrid.counter += 1
        self.name = "microgrid{:d}".format(Microgrid.counter)
        self.mode = mode

        # Components
        self.buses = list()
        self.ev_parks = list()
        self.batteries = list()
        self.productions = list()
        self.lines = list()
        self.sensors = list()
        self.circuitbreakers = list()
        self.disconnectors = list()
        self.intelligent_switches = list()
        self.controller = MicrogridController(
            name=self.name + "_controller",
            power_network=self,
        )
        self.comp_list = list()
        self.comp_dict = dict()

        # PowerNetwork connections
        self.distribution_network = distribution_network
        self.distribution_network.add_child_network(self)
        self.distribution_network.power_system.controller.add_microgrid_controller(
            self.controller
        )
        self.distribution_network.power_system.comp_dict[
            self.controller.name
        ] = self.controller
        self.distribution_network.power_system.comp_list.append(
            self.controller
        )

        self.failed_line = False

        self.connected_line = None
        self.circuitbreaker = None
        self.add_connected_line(connected_line, mode)

        # Load shedding
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0
        # Sectioning
        self.sections = None
        ## History
        self.history: dict = {}
        self.monte_carlo_history: dict = {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Microgrid(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def add_connected_line(self, connected_line, mode):
        """
        Sets the line connecting the microgrid to overlying network

        Parameters
        ----------
        connected_line : Line
            The line connecting the distribution system to overlaying network
        mode : str
            Which mode the microgrid follows

        Returns
        ----------
        None

        """
        # Sets the connected line for the microgrid
        self.connected_line = connected_line

        # Add the components attached to the line
        # to the microgrid:

        # CircuitBreaker
        self.circuitbreaker = connected_line.circuitbreaker
        self.circuitbreaker.mode = mode
        if self.circuitbreaker is None:
            raise Exception(
                "{} connects the microgrid to the "
                "distribution network, it must contain a circuitbreaker".format(
                    connected_line
                )
            )
        self.comp_dict[self.circuitbreaker.name] = self.circuitbreaker
        self.comp_list.append(self.circuitbreaker)

        # Line
        self.add_line(connected_line)

    def add_bus(self, bus: Bus):
        """
        Adding a bus including elements on the bus (battery, generation unit, EV park) to the microgrid

        Parameters
        ----------
        bus : Bus
            A Bus element

        Returns
        ----------
        None

        """

        # Add bus to microgrid
        self.comp_dict[bus.name] = bus
        bus.handle.color = self.color
        bus.color = self.color
        bus.parent_network = self
        self.buses.append(bus)
        self.buses = unique(self.buses)

        # Add components attached to bus to microgrid:

        # EV-Park
        if bus.ev_park is not None:
            self.comp_dict[bus.ev_park.name] = bus.ev_park
            self.comp_list.append(bus.ev_park)
            self.ev_parks.append(bus.ev_park)
            self.ev_parks = unique(self.ev_parks)

        # Battery
        if bus.battery is not None:
            # Set microgrid mode
            bus.battery.set_mode(self.mode)
            self.comp_dict[bus.battery.name] = bus.battery
            self.comp_list.append(bus.battery)
            self.batteries.append(bus.battery)
            self.batteries = unique(self.batteries)

        # Production
        if bus.prod is not None:
            self.comp_dict[bus.prod.name] = bus.prod
            self.comp_list.append(bus.prod)
            self.productions.append(bus.prod)
            self.productions = unique(self.productions)

        # Add bus to the power system
        self.distribution_network.power_system.add_bus(bus)

    def add_buses(self, buses: list):
        """
        Adding buses to the microgrid

        Parameters
        ----------
        buses : list
            A list of Bus elements in the microgrid

        Returns
        ----------
        None

        """
        for bus in buses:
            self.add_bus(bus)

    def add_line(self, line: Line):
        """
        Adding a line and the components connected to the line to the microgrid

        Parameters
        ----------
        line : Line
            A Line element

        Returns
        ----------
        None

        """

        # Add line to microgrid
        line.handle.color = self.color
        line.color = self.color
        self.lines.append(line)
        self.lines = unique(self.lines)

        # Add components attached to bus to microgrid:

        # Sensor
        if line.sensor:
            self.comp_dict[line.sensor.name] = line.sensor
            self.comp_list.append(line.sensor)
            self.sensors.append(line.sensor)
            self.sensors = unique(self.sensors)
            self.controller.sensors.append(line.sensor)
            self.controller.sensors = unique(self.controller.sensors)

        # Disconnector
        for discon in line.disconnectors:
            self.comp_dict[discon.name] = discon
            self.comp_list.append(discon)
            self.disconnectors.append(discon)
            self.disconnectors = unique(self.disconnectors)

            # Intelligent switch
            if discon.intelligent_switch:
                self.comp_dict[
                    discon.intelligent_switch.name
                ] = discon.intelligent_switch
                self.comp_list.append(discon.intelligent_switch)
                self.intelligent_switches.append(discon.intelligent_switch)
                self.intelligent_switches = unique(self.intelligent_switches)

        # Set microgrid as parent network
        line.add_parent_network(self)

        # Add line to power system
        self.distribution_network.power_system.add_line(line)

    def add_lines(self, lines: list):
        """
        Adding lines to the microgrid

        Parameters
        ----------
        lines : list
            A list of Line elements in the microgrid

        Returns
        ----------
        None

        """
        for line in lines:
            self.add_line(line)

    def get_lines(self):
        """
        Returns the lines in the microgrid

        Parameters
        ----------
        None

        Returns
        ----------
        lines : list
            List of Line elements

        """
        return self.lines

    def connect(self):
        """
        Connects the microgrid to the parent distribution network by closing the disconnectors on the microgrid lines


        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for discon in self.connected_line.get_disconnectors():
            if discon.is_open:
                discon.close()

    def disconnect(self):
        """
        Disconnects the microgrid from the parent distribution network by opening the disconnectors on the microgrid lines

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for discon in self.connected_line.get_disconnectors():
            if not discon.is_open:
                discon.open()

    def reset_slack_bus(self):
        """
        Resets the slack bus attribute of the buses in the microgrid

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for bus in self.buses:
            bus.is_slack = False

    def get_max_load(self):
        """
        Get the maximum load of the microgrid for the entire load history and returns the max load

        Parameters
        ----------
        None

        Returns
        ----------
        p_load_max : float
            The maximum active power load of the microgrid for the entire loading history

        q_load_max : float
            The maximum reactive power load of the microgrid for the entire loading history

        """
        p_load_max, q_load_max = 0, 0
        for bus in self.buses:
            if bus.pload_data != list():
                d_bus = bus  # Dummy bus used to find number of increments
                n_increments = len(d_bus.pload_data[0])  # Number of increments
                break
        for increment in range(n_increments):
            p_load, q_load = 0, 0
            for bus in self.buses:
                for i in range(len(bus.pload_data)):
                    p_load += bus.pload_data[i][increment] * bus.n_customers
                    q_load += bus.qload_data[i][increment] * bus.n_customers
            p_load_max = max(p_load_max, p_load)
            q_load_max = max(q_load_max, q_load)
        return p_load_max, q_load_max

    def get_monte_carlo_history(self, attribute):
        """
        Returns the specified history variable from the Monte Carlo simulation

        Parameters
        ----------
        attribute : str
            Microgrid attribute

        Returns
        ----------
        monte_carlo_history[attribute] : dict
            Returns the history variables of an attribute from the Monte Carlo simulation

        """
        return self.monte_carlo_history[attribute]

    def get_history(self, attribute):
        """
        Returns the specified history variable

        Parameters
        ----------
        attribute : str
            Microgrid attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def get_system_load(self):
        """
        Returns the system load in the microgrid at the current time in MW and MVar

        Parameters
        ----------
        None

        Returns
        ----------
        pload : float
            The active power load in the microgrid
        qload : float
            The reactive power load in the microgrid

        """
        pload, qload = 0, 0
        for bus in self.buses:
            p, q = bus.get_load()
            pload += p
            qload += q
        return pload, qload

    def reset_energy_shed_variables(self):
        """
        Resets the energy.shed variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0



from abc import ABC, abstractmethod

from .Line import Line


class Switch(ABC):

    # Enforce line-list as attribute
    line: Line = None

    @abstractmethod
    def open(self):
        pass

    @abstractmethod
    def close(self):
        pass



from .monte_carlo import save_monte_carlo_history_from_dict

from .sequence import save_history

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



from enum import Enum

from relsad.network.components import Controller, Disconnector
from relsad.Time import Time
from relsad.topology.ICT.dfs import is_connected


class SectionState(Enum):
    """
    Section state

    Attributes
    ----------
    CONNECTED : int
        The section is connected to its parent power network
    DISCONNECTED : int
        The section is disconnected from its parent power network
    """

    CONNECTED = 1
    DISCONNECTED = 2


class Section:

    """
    Common class for sections

    ...

    Attributes
    ----------
    lines : list
        List of lines in the section
    switches : list
        List of switches in the section
    parent_section: Section
        The parent section for the section
    child_sections : list
        List of child sections
    state : SectionState
        The state of the section

    Methods
    ----------
    add_child_section(section)
        Adds child section to section
    attach_to_lines
        Adds lines to section and child section
    connect(dt)
        Connects the sections, connects the lines in the section and closes
        the switches in the section
    connect_manually()
        Connects the sections, connects the lines in the section and closes
        the switches in the section. Used when no ICT
    get_disconnect_time(dt)
        Returns the total outage time (the time the section is disconnected)
        of the section
    disconnect()
        Disconnects the section, the lines in the section, and opens
        the switches in the section



    """

    __slots__ = (
        "rank",
        "lines",
        "switches",
        "parent_section",
        "child_sections",
        "state",
    )

    def __init__(
        self,
        parent_section,
        lines: list,
        switches: list,
    ):
        self.lines = lines
        self.switches = switches
        self.parent_section = parent_section
        self.child_sections = []
        self.state = SectionState.CONNECTED

    def __str__(self):
        return str(self.lines)

    def __repr__(self):
        return f"Section({str(self.lines)})"

    def __eq__(self, other):
        if hasattr(other, "lines"):
            return set(self.lines) - set(other.lines) == set()
        else:
            return False

    def __hash__(self):
        return hash(str(self.lines))

    def add_child_section(self, section):
        """
        Adds child section to section

        Parameters
        ----------
        section : Section
            A Section element

        Returns
        -------
        None

        """
        if section not in self.child_sections:
            self.child_sections.append(section)

    def attach_to_lines(self):
        """
        Adds lines to section and child section

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        for line in self.lines:
            line.section = self
        for child_section in self.child_sections:
            child_section.attach_to_lines()

    def connect(self, dt: Time, controller: Controller):
        """
        Connects the sections, connects the lines in the section and closes
        the switches in the section

        Parameters
        ----------
        dt : Time
            The current time step
        controller : Controller
            The power network controller

        Returns
        ----------
        None

        """
        self.state = SectionState.CONNECTED
        for line in self.lines:
            line.connect()
        for switch in self.switches:
            # Skip if switch is a CircuitBreaker
            if not isinstance(switch, type(Disconnector)):
                break
            # If no ICT network
            if controller.ict_node is None:
                switch.intelligent_switch.close(dt)
            # If both components have ICT nodes
            elif (
                controller.ict_node is not None
                and switch.intelligent_switch.ict_node is not None
            ):
                # If the ICT nodes are connected to each other
                if is_connected(
                    node_1=controller.ict_node,
                    node_2=switch.intelligent_switch.ict_node,
                    network=controller.ict_network,
                ):
                    switch.intelligent_switch.close(dt)
                # If the ICT nodes are not connected to each other
                else:
                    # Repair crew is assumed to be present repairing the line,
                    # they close the switch manually
                    switch.close()
            # If no ICT node on intelligent switch
            else:
                # Repair crew is assumed to be present repairing the line,
                # they close the switch manually
                switch.close()

    def connect_manually(self):
        """
        Connects the sections, connects the lines in the section and closes
        the switches in the section. Used when no ICT

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = SectionState.CONNECTED
        for line in self.lines:
            line.connect()
        for switch in self.switches:
            switch.close()

    def get_disconnect_time(self, dt: Time, controller: Controller):
        """
        Returns the total outage time (the time the section is disconnected)
        of the section

        Parameters
        ----------
        dt : Time
            The current time step
        controller : Controller
            The power network controller

        Returns
        ----------
        sectioning_time : Time
            The sectioning time of the section

        """
        sectioning_time = Time(0)
        need_manual_attention = False
        for switch in self.switches:
            # Skip if switch is a CircuitBreaker
            if not isinstance(switch, type(Disconnector)):
                break
            # If no ICT network
            if controller.ict_node is None:
                sectioning_time += switch.intelligent_switch.get_open_time(dt)
            # If both components have ICT nodes
            elif (
                controller.ict_node is not None
                and switch.intelligent_switch.ict_node is not None
            ):
                # If the ICT nodes are connected to each other
                if is_connected(
                    node_1=controller.ict_node,
                    node_2=switch.intelligent_switch.ict_node,
                    network=controller.ict_network,
                ):
                    sectioning_time += switch.intelligent_switch.get_open_time(
                        dt
                    )
                # If the ICT nodes are not connected to each other
                else:
                    need_manual_attention = True
            # If no ICT node on intelligent switch
            else:
                need_manual_attention = True
        if need_manual_attention is True:
            sectioning_time += controller.manual_sectioning_time
        for line in self.lines:
            line.remaining_outage_time += sectioning_time
        return sectioning_time

    def disconnect(self):
        """
        Disconnects the section, the lines in the section, and opens
        the switches in the section

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = SectionState.DISCONNECTED
        for line in self.lines:
            line.disconnect()
        for switch in self.switches:
            switch.open()



"""
This module contains set operation utilities
"""


def unique(sequence: list):
    """
    Return list of unique elements while preserving the order

    Parameters
    ----------
    sequence : list
        List of elements

    Returns
    -------
    unq : list
        List of unique elements with preserved order

    """
    seen = set()
    unq = [x for x in sequence if not (x in seen or seen.add(x))]
    return unq


def subtract(list1: list, list2: list):
    """
    Return difference between lists while preserving the order

    Parameters
    ----------
    list1 : list
    list2 : list

    Returns
    -------
    difference : list
        The difference between lists with preserved order

    """

    difference = [x for x in list1 if x not in list2]
    return difference


def intersection(list1: list, list2: list):
    """
    Returns the intersection between two list while preserving the order

    Parameters
    ----------
    list1 : list
        List 1
    list2 : list
        List 2

    Returns
    -------
    intersec : list
        The intersection between two list with preserved order

    """
    intersec = [x for x in list1 if x in list2]

    return intersec



import matplotlib.lines as mlines
import numpy as np

from relsad.Table import Table
from relsad.Time import Time, TimeUnit

from .Battery import Battery, BatteryType
from .Bus import Bus
from .Component import Component
from .MicrogridController import MicrogridMode


class EVPark(Component):

    """
    Common class for batteries
    ...

    Attributes
    ----------
    name : string
        Name of the EV park
    bus : Bus
        The bus the EV park is connected to
    num_ev_dist : Table
        Statistical distribution that returns the amount of EVs available at a given time in the network
    inj_p_max : float
        The active power charging/discharging capacity of the EV battery [MW]
    inj_q_max : float
        The reactive power discharging capacity of the EV battery [MVar]
    E_max : float
        The maximum energy capacity of an EV [Mwh]
    SOC_min : float
        The minimal state of charge level in the EV battery
    SOC_max : float
        The maximum state of charge level in the EV battery
    n_battery : float
        The EV battery efficiency
    v2g_flag : bool
        A flag telling if the EV park contributes with V2G services or not
    curr_p_demand : float
        Current demand of active power from the EV park [MW]
    curr_q_demand : float
        Current demand of reactive power from the EV park [MVar]
    curr_p_charge : float
        Current active power that are being charged/dischared [MW]
    curr_q_charge : float
        Current reactive power that are being charged/dischared [MVar]
    available_cars : list
        List of available cars in the EV park
    available_num_cars : int
        Number of available cars in the EV park
    num_cars : int
        Number of cars in the EV park
    num_consecutive_interruptions : int
        Number of consecutive interruptions experienced by the EV park
    park_interruption_fraction : float
        Fraction of interruption experienced by the EV park
    acc_available_num_cars : int
        Accumulated available number of cars in EV park
    acc_num_interruptions : float
        Accumulated number of interruptions experienced by the EV park
    acc_exp_interruptions : float
        Accumulated experienced interruptions in the EV park
    curr_exp_interruptions : float
        Current experienced interruptions in the EV park
    acc_exp_car_interruptions : float
        Accumulated experienced car interruptions in the EV park
    curr_exp_car_interruptions : float
        Current experienced car interruptions in the EV park
    curr_interruption_duration : Time
        Current interruption duration experienced by the EV park
    acc_interruption_duration : Time
        Accumulated interruption duration experienced by the EV park
    history : dict
        Dictonary attribute that stores the historic variables

    Methods
    ----------
    draw_current_state(hour_of_day)
        Draws the number of EVs in the park at that time and the SOC level of each EV which will make the SOC level of the EV park
    update(p, q, fail_duration, dt, hour_of_day)
        Updates the EV park status for the current time step
    get_curr_demand(dt)
        Returns the current power demand of an EV
    get_SOC()
        Returns the SOC level of the EV park
    get_ev_index()
        Returns the power demand of av EV that is not met by the system
    initialize_history()
        Initializes the history variables
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    update_fail_status(dt)
        Locks og unlocks the battery functionality based on failure states of the basestation
    add_random_instance(random_gen)
        Adds global random seed
    print_status()
        Prints the status of the EV park
    reset_status(save_flag)
        Resets and sets the status of the class parameters

    """

    ## Visual attributes
    marker = "x"
    size = 3**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=1,
        markersize=size,
        linestyle="None",
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        bus: Bus,
        num_ev_dist: Table,
        inj_p_max: float = 0.072,
        inj_q_max: float = 0.072,
        E_max: float = 0.70,
        SOC_min: float = 0.2,
        SOC_max: float = 0.9,
        n_battery: float = 0.95,
        v2g_flag: bool = True,
    ):

        # Verify input
        if bus is None:
            raise Exception("EVPark must be connected to a Bus")
        if bus.parent_network is not None:
            raise Exception(
                "EVPark must be created before the bus is connected to a network"
            )
        if num_ev_dist is None:
            raise Exception(
                "EVPark must have a table distributing the number"
                "of EV during the day"
            )
        if inj_p_max < 0:
            raise Exception("The active power injection must be positive")
        if inj_q_max < 0:
            raise Exception("The reactive power injection must be positive")
        if E_max < 0:
            raise Exception("The energy capacity must be positive")
        if SOC_min < 0 or SOC_max > 1:
            raise Exception("The SOC limits must be between 0 and 1")
        if n_battery < 0 or n_battery > 1:
            raise Exception("The efficiency must be between 0 and 1")

        self.name = name

        self.bus = bus
        bus.ev_park = self

        self.num_ev_dist = num_ev_dist
        self.inj_p_max = inj_p_max  # MW
        self.inj_q_max = inj_q_max  # MVar
        self.E_max = E_max  # MWh
        self.SOC_min = SOC_min
        self.SOC_max = SOC_max
        self.n_battery = n_battery

        self.curr_p_demand = 0  # MW
        self.curr_q_demand = 0  # MVar
        # curr_p_charge < 0 => discharge
        # curr_p_charge > 0 => charge
        self.curr_p_charge = 0  # MW
        self.curr_q_charge = 0  # MVar

        self.v2g_flag = v2g_flag

        self.available_cars = list()
        self.available_num_cars = 0
        self.num_cars = round(max(self.num_ev_dist.y))

        ## Reliability attributes
        self.num_consecutive_interruptions = 0
        self.park_interruption_fraction = 0
        self.acc_available_num_cars = 0
        self.acc_num_interruptions = 0
        self.curr_exp_interruptions = 0
        self.acc_exp_interruptions = 0
        self.curr_exp_car_interruptions = 0
        self.acc_exp_car_interruptions = 0
        self.curr_interruption_duration = Time(0)
        self.acc_interruption_duration = Time(0)

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"EVPark(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, EVPark)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def draw_current_state(self, hour_of_day: int):
        """
        Draws the number of EVs in the park at that time
        and the SOC level of each EV which will make the SOC level of the EV park

        Parameters
        ----------
        hour_of_day : int
            The hour of the day
        Returns
        ----------
        None

        """
        # Draw number of cars in the EV park
        self.available_num_cars = round(
            self.num_ev_dist.get_value(hour_of_day)
        )

        # Draw the SOC states of the cars in the EV park
        soc_states = self.ps_random.uniform(
            low=self.SOC_min,
            high=self.SOC_max,
            size=self.available_num_cars,
        )

        # Create the current cars in the EV park with
        # their respective status
        self.available_cars = [
            Battery(
                name="ev{:d}_{}".format(i, self.name),
                bus=self.bus,
                inj_p_max=self.inj_p_max,
                inj_q_max=self.inj_q_max,
                E_max=self.E_max,
                SOC_min=self.SOC_min,
                SOC_max=self.SOC_max,
                n_battery=self.n_battery,
                battery_type=BatteryType.EV,
                random_instance=self.ps_random,
                SOC_start=soc_states[i],
            )
            for i in range(self.available_num_cars)
        ]

    def update(
        self,
        p: float,
        q: float,
        fail_duration: Time,
        dt: Time,
        hour_of_day: int,
    ):
        """
        Updates the EV park status for the current time step

        Parameters
        ----------
        p : float
            Active power balance of the parent power system
        q : float
            Reactive power balance of the parent power system
        fail_duration : Time
            The duration of the current failure
        dt : Time
            The current time step
        hour_of_day : int
            The hour of the day
        Returns
        ----------
        p : float
            Remaining active power balance of the parent power system
        q : float
            Remaining reactive power balance of the parent power system

        """
        if fail_duration == dt:
            self.draw_current_state(hour_of_day)
            self.acc_available_num_cars += self.available_num_cars

        # Update car batteries based on system balance
        p_start = p
        q_start = q
        for car in self.available_cars:
            if self.v2g_flag is True:
                # Vehicle to grid: allow charge and discharge
                p, q = car.update_bus_load_and_prod(p, q, dt)
            else:
                # Not vehicle to grid: allow charge only
                if p < 0:
                    p, q = car.update_bus_load_and_prod(p, q, dt)
        p_change = p_start - p
        q_change = q_start - q

        # Define load or production generated from the EV park
        pprod = max(0, p_change)
        qprod = max(0, q_change)
        pload = abs(min(0, p_change))
        qload = abs(min(0, q_change))

        # Update current demand and charge attributes
        (
            self.curr_p_demand,
            self.curr_q_demand,
        ) = self.get_curr_demand(dt)
        self.curr_p_charge = pload - pprod
        self.curr_q_charge = qload - qprod
        return p, q

    def get_curr_demand(self, dt: Time):
        """
        Returns the current power demand of an EV

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        curr_p_demand : float
            The current active power demand of the EV
        curr_p_demand : float
            The current reactive power demand of the EV

        """
        curr_p_demand = 0
        curr_q_demand = 0
        for car in self.available_cars:
            curr_p_demand += min(
                car.inj_p_max,
                (car.E_max * car.SOC_max - car.E_battery) / dt.get_hours(),
            )
        return curr_p_demand, curr_q_demand

    def get_SOC(self):
        """
        Returns the SOC level of the EV park

        Parameters
        ----------
        None

        Returns
        ----------
        mean_SOC: float
            The average SOC value of the EV park

        """

        if self.available_num_cars <= 0:
            return 0
        mean_SOC = np.mean([car.SOC for car in self.available_cars])
        return mean_SOC

    def get_ev_index(self):
        """
        Returns the power demand of an EV that is not met by the system

        Parameters
        ----------
        None

        Returns
        ----------
        ev_index: float
            The power demand of an EV that is not met by the system

        """
        ev_index = self.curr_p_demand - self.curr_p_charge
        return ev_index

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None
        """

        self.history["SOC"] = {}
        self.history["ev_index"] = {}
        self.history["demand"] = {}
        self.history["charge"] = {}
        self.history["num_cars"] = {}
        self.history["available_num_cars"] = {}
        self.history["park_interruption_fraction"] = {}
        self.history["acc_num_interruptions"] = {}
        self.history["acc_exp_interruptions"] = {}
        self.history["acc_exp_car_interruptions"] = {}
        self.history["acc_interruption_duration"] = {}
        self.history["acc_available_num_cars"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None
        """
        dt = curr_time - prev_time if prev_time is not None else curr_time
        # Accumulate fraction of interupted customers
        self.curr_exp_car_interruptions = (
            abs(self.curr_p_charge / self.inj_p_max)
            if self.curr_p_charge < 0
            else 0
        )
        self.park_interruption_fraction = (
            self.curr_exp_car_interruptions / self.num_cars
        )

        if self.park_interruption_fraction > 0:
            self.curr_exp_interruptions += self.park_interruption_fraction
            self.num_consecutive_interruptions += 1
            self.curr_interruption_duration += dt
        else:
            if self.num_consecutive_interruptions >= 1:
                self.acc_num_interruptions += 1
                self.acc_exp_interruptions += (
                    self.curr_exp_interruptions
                    / self.num_consecutive_interruptions
                )
                self.acc_exp_car_interruptions += (
                    self.curr_exp_car_interruptions
                    / self.num_consecutive_interruptions
                )
                self.acc_interruption_duration += (
                    self.curr_interruption_duration
                )
            self.curr_exp_interruptions = 0
            self.curr_exp_car_interruptions = 0
            self.curr_interruption_duration = Time(0)
            self.num_consecutive_interruptions = 0
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["SOC"][time] = self.get_SOC()
            self.history["ev_index"][time] = self.get_ev_index()
            self.history["demand"][time] = self.curr_p_demand
            self.history["charge"][time] = self.curr_p_charge
            self.history["num_cars"][time] = self.num_cars
            self.history["available_num_cars"][
                time
            ] = self.available_num_cars
            self.history["park_interruption_fraction"][
                time
            ] = self.park_interruption_fraction
            self.history["acc_num_interruptions"][
                time
            ] = self.acc_num_interruptions
            self.history["acc_exp_interruptions"][
                time
            ] = self.acc_exp_interruptions
            self.history["acc_exp_car_interruptions"][
                time
            ] = self.acc_exp_car_interruptions
            self.history["acc_interruption_duration"][
                time
            ] = self.acc_interruption_duration.get_hours()
            self.history["acc_available_num_cars"][
                time
            ] = self.acc_available_num_cars

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def update_fail_status(self, dt: Time):
        """
        Locks og unlocks the battery functionality based on failure states of the basestation

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        pass

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):

        """
        Prints the status of the EV park

        Parameters
        ----------
        None

        Returns
        ----------
        None
        """

        print("EV park status")
        print("name: {}".format(self.name))
        print("parent bus: {}".format(self.bus.name))
        print("inj_p_max: {} MW".format(self.inj_p_max))
        print("inj_q_max: {} MVar".format(self.inj_q_max))
        print("E_max: {} MWh".format(self.E_max))
        print("SOC_min: {}".format(self.SOC_min))
        print("SOC_max: {}".format(self.SOC_max))
        print("SOC: {:.2f}".format(self.get_SOC))

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.curr_p_demand = 0
        self.curr_q_demand = 0
        self.curr_p_charge = 0
        self.curr_q_charge = 0
        self.available_cars.clear()
        self.available_num_cars = 0
        ## Reliability attributes
        self.num_consecutive_interruptions = 0
        self.park_interruption_fraction = 0
        self.acc_available_num_cars = 0
        self.acc_num_interruptions = 0
        self.curr_exp_interruptions = 0
        self.acc_exp_interruptions = 0
        self.curr_exp_car_interruptions = 0
        self.acc_exp_car_interruptions = 0
        self.acc_interruption_duration = Time(0)
        self.curr_interruption_duration = Time(0)
        if save_flag:
            self.initialize_history()



import copy
from enum import Enum
import os
from multiprocessing import Pool

import numpy as np
from numpy.random import SeedSequence

from relsad.energy.shedding import shed_energy
from relsad.loadflow.ac.bfs import run_bfs_load_flow
from relsad.network.systems import PowerNetwork, PowerSystem
from .monte_carlo.history import (
    merge_monte_carlo_history,
    save_network_monte_carlo_history,
)
from .sequence.history import save_sequence_history
from .system_config import (
    find_sub_systems,
    prepare_system,
    reset_system,
    update_sub_system_slack,
)
from relsad.Time import Time, TimeStamp, TimeUnit


class Simulation:
    """
    Common class for simulation

    ...

    Attributes
    ----------
    power_system : PowerSystem
        A PowerSystem element
    random_seed : int
        Random seed number
    fail_duration : Time
        The duration of a failure

    Methods
    ----------
    distribute_random_instance(random_intance)
        Adds a global numpy random instance
    run_load_flow(network)
        Runs load flow of a network
    run_increment(inc_idx, start_time, prev_time, curr_time, save_flag)
        Runs power system at current state for on time increment
    run_sequence(start_time, time_array, time_unit, save_flag)
        Runs power system for a sequence of increments
    run_sequential(start_time, stop_time, time_step, time_unit, save_dir, save_flag)
        Runs a sequential simulation with the power system
    run_iteration(it, start_time, time_array, time_unit, save_dir, save_iterations, random_seed)
        Runs a sequential iteration with the power system
    run_monte_carlo(iterations, start_time, stop_time, time_step, time_unit, save_iterations, save_dir, n_procs, debug)
        Runs Monte Carlo simulation of the power system
    """

    def __init__(self, power_system: PowerSystem, random_seed: int = None):
        self.power_system = power_system
        self.power_system.verify_component_setup()
        self.random_seed = random_seed
        self.fail_duration = Time(0)

    def distribute_random_instance(self, random_instance):
        """
        Adds a global numpy random instance

        Parameters
        ----------
        random_instance : np.random.Generator
            A random generator

        Returns
        ----------
        None

        """
        self.power_system.random_instance = random_instance
        for comp in self.power_system.comp_list:
            comp.add_random_instance(random_instance)
        self.power_system.controller.add_random_instance(random_instance)

    def run_load_flow(self, network: PowerNetwork):
        """
        Runs load flow of a network

        Parameters
        ----------
        network : PowerNetwork
            A PowerNetwork element

        Returns
        ----------
        None

        """
        run_bfs_load_flow(network)

    def run_increment(
        self,
        inc_idx: int,
        start_time: TimeStamp,
        prev_time: Time,
        curr_time: Time,
        save_flag: bool = True,
    ):
        """
        Runs power system at current state for one time increment

        Parameters
        ----------
        inc_idx : int
            Increment index
        start_time : TimeStamp
            The start time of the simulation/iteration
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        ## Time step
        dt = curr_time - prev_time if prev_time is not None else curr_time
        ## Set loads
        self.power_system.set_load_and_cost(inc_idx=inc_idx)
        ## Set productions
        self.power_system.set_prod(inc_idx=inc_idx)
        ## Set fail status
        self.power_system.update_fail_status(dt=dt)
        ## Run control loop
        self.power_system.controller.run_control_loop(
            curr_time=curr_time,
            dt=dt,
        )
        if (
            self.power_system.failed_comp()
            or not self.power_system.full_batteries()
        ):
            self.fail_duration += dt
            ## Find sub systems
            find_sub_systems(
                p_s=self.power_system,
                curr_time=curr_time,
            )
            update_sub_system_slack(p_s=self.power_system)
            ## Load flow
            for sub_system in self.power_system.sub_systems:
                ## Update batteries and history
                sub_system.update_batteries(
                    fail_duration=self.fail_duration,
                    dt=dt,
                )
                ## Update EV parks
                sub_system.update_ev_parks(
                    fail_duration=self.fail_duration,
                    dt=dt,
                    start_time=start_time,
                    curr_time=curr_time,
                )
                ## Run load flow
                sub_system.reset_load_flow_data()
                if sub_system.slack is not None:
                    self.run_load_flow(network=sub_system)
                ## Shed load
                shed_energy(
                    power_system=sub_system,
                    dt=dt,
                )
            ## Log results
            self.power_system.update_sequence_history(
                prev_time=prev_time,
                curr_time=curr_time,
                save_flag=save_flag,
            )
            self.power_system.reset_load_flow_data()
        else:
            if self.fail_duration > Time(0):
                ## Log results
                self.power_system.update_sequence_history(
                    prev_time=prev_time,
                    curr_time=curr_time,
                    save_flag=save_flag,
                )
                ## Reset fail duration
                self.fail_duration = Time(0)

    def run_sequence(
        self,
        start_time: TimeStamp,
        time_array: np.ndarray,
        time_unit: TimeUnit,
        callback: callable = None,
        save_flag: bool = True,
    ):
        """
        Runs power system for a sequence of increments

        Parameters
        ----------
        start_time : TimeStamp
            The start time of the simulation/iteration
        time_array : np.ndarray
            Time array
        time_unit : TimeUnit
            Time unit
        callback : callable, optional
            A callback function that allows for user-defined
            behavior. The callback function is called at the start
            of every increment. The callback function must contain
            the following arguments: ps, prev_time, curr_time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        prev_time = Time(0, unit=time_unit)
        curr_time = Time(0, unit=time_unit)
        for inc_idx, time_quantity in enumerate(time_array):
            curr_time = Time(time_quantity, unit=time_unit)
            if callback is not None:
                callback(
                    ps=self.power_system,
                    prev_time=prev_time,
                    curr_time=curr_time,
                )
            self.run_increment(
                inc_idx,
                start_time,
                prev_time,
                curr_time,
                save_flag,
            )
            prev_time = copy.deepcopy(curr_time)

    def run_sequential(
        self,
        start_time: TimeStamp,
        stop_time: TimeStamp,
        time_step: Time,
        time_unit: TimeUnit,
        callback: callable = None,
        save_dir: str = "results",
        save_flag: bool = True,
    ):
        """
        Runs sequential simulation of the power system

        Parameters
        ----------
        start_time : TimeStamp
            The start time of the simulation/iteration
        stop_time : TimeStamp
            The stop time of the simulation/iteration
        time_step : Time
            A time step (1 hour, 2 hours, ect.)
        time_unit : TimeUnit
            A time unit (hour, seconds, ect.)
        callback : callable, optional
            A callback function that allows for user-defined
            behavior. The callback function is called at the start
            of every increment. The callback function must contain
            the following arguments: ps, prev_time, curr_time
        save_dir : str
            The saving directory
        save_flag : bool
            Flag for saving the simulation results

        Returns
        ----------
        None

        """
        if callback is not None and not callable(callback):
            raise Exception("The callback argument must be callable")

        # Initialize random seed
        ss = SeedSequence(self.random_seed)
        random_seed = ss.spawn(1)[0]

        # Initiate random instance
        random_instance = np.random.default_rng(random_seed)

        # Distribute random instance to power system components
        self.distribute_random_instance(random_instance)

        # Prepare power system for simulation
        time_array = prepare_system(
            power_system=self.power_system,
            start_time=start_time,
            stop_time=stop_time,
            time_step=time_step,
            time_unit=time_unit,
        )

        # Initialize sequence history variables
        self.power_system.initialize_sequence_history()

        # Run sequence
        self.run_sequence(
            start_time=start_time,
            time_array=time_array,
            time_unit=time_unit,
            callback=callback,
            save_flag=save_flag,
        )

        if save_flag is True:
            # Save sequence history
            save_sequence_history(
                power_system=self.power_system,
                time_unit=time_unit,
                save_dir=os.path.join(
                    save_dir,
                    "sequence",
                ),
            )

    def run_iteration(
        self,
        it: int,
        start_time: TimeStamp,
        time_array: np.array,
        time_unit: TimeUnit,
        save_dir: str,
        save_flag: bool,
        random_seed: int,
        callback: callable = None,
    ):
        """
        Runs power system for an iteration

        Parameters
        ----------
        it : int
            Iteration number
        start_time : TimeStamp
            The start time of the simulation/iteration
        stop_time : TimeStamp
            The stop time of the simulation/iteration
        time_array : np.array
            A time step (1 hour, 2 hours, ect.)
        time_unit : TimeUnit
            A time unit (hour, seconds, ect.)
        save_dir: str
            The saving directory
        save_flag : bool
            Flag for saving the iteration results
        random_seed : int
            Random seed number
        callback : callable, optional
            A callback function that allows for user-defined
            behavior. The callback function is called at the start
            of every increment. The callback function must contain
            the following arguments: prev_time, curr_time

        Returns
        ----------
        save_dict : dict
            Dictionary with simulation results

        """
        # Initiate random instance
        if random_seed is None:
            random_instance = np.random.default_rng()
        else:
            random_instance = np.random.default_rng(random_seed)

        # Distribute random instance to power system components
        self.distribute_random_instance(random_instance)

        # Initialize monte carlo history variables
        save_dict = self.power_system.initialize_monte_carlo_history()

        # Print current iteration
        print(f"it: {it}", flush=True)

        # Reset power system
        reset_system(self.power_system, save_flag)

        # Initialize sequence history variables
        self.power_system.initialize_sequence_history()

        # Run iteration sequence
        self.run_sequence(
            start_time=start_time,
            time_array=time_array,
            time_unit=time_unit,
            callback=callback,
            save_flag=save_flag,
        )

        if save_flag is True:
            # Save sequence history
            save_sequence_history(
                power_system=self.power_system,
                time_unit=time_unit,
                save_dir=os.path.join(
                    save_dir,
                    "sequence",
                    str(it),
                ),
            )

        # Update monte carlo history variables
        sim_duration = Time(time_array[-1], time_unit)
        save_dict = self.power_system.update_monte_carlo_history(
            it=it,
            current_time=sim_duration,
            save_dict=save_dict,
        )
        return save_dict

    def run_monte_carlo(
        self,
        iterations: int,
        start_time: TimeStamp,
        stop_time: TimeStamp,
        time_step: Time,
        time_unit: TimeUnit,
        callback: callable = None,
        save_iterations: list = [],
        save_dir: str = "results",
        n_procs: int = 1,
        debug: bool = False,
        save_flag: bool = True,
    ):
        """
        Runs Monte Carlo simulation of the power system

        Parameters
        ----------
        iterations : int
            Number of iterations
        start_time : TimeStamp
            The start time of the simulation/iteration
        stop_time : TimeStamp
            The stop time of the simulation/iteration
        time_step : Time
            A time step (1 hour, 2 hours, ect.)
        time_unit : TimeUnit
            A time unit (hour, seconds, ect.)
        callback : callable, optional
            A callback function that allows for user-defined
            behavior. The callback function is called at the start
            of every increment. The callback function must contain
            the following arguments: prev_time, curr_time
        save_iterations : list
            List of iterations where the sequence results will be saved
        save_dir : str
            The saving directory
        n_procs : int
            Number of processors
        debug : bool
            Indicates if debug mode is on or off
        save_flag : bool
            Flag for saving the simulation results

        Returns
        ----------
        None

        """
        if callback is not None and not callable(callback):
            raise Exception("The callback argument must be callable")

        # Initialize random seeds
        ss = SeedSequence(self.random_seed)
        child_seeds = ss.spawn(iterations)

        # Prepare power system for simulation
        time_array = prepare_system(
            power_system=self.power_system,
            start_time=start_time,
            stop_time=stop_time,
            time_step=time_step,
            time_unit=time_unit,
        )

        # Run iterations
        if debug:
            it_dicts = []
            for it in range(1, iterations + 1):
                it_dicts.append(
                    self.run_iteration(
                        it=it,
                        start_time=start_time,
                        time_array=time_array,
                        time_unit=time_unit,
                        save_dir=save_dir,
                        save_flag=(
                            it in save_iterations and save_flag is True
                        ),
                        random_seed=child_seeds[it - 1],
                        callback=callback,
                    )
                )
        else:
            with Pool(processes=n_procs) as pool:
                it_dicts = pool.starmap(
                    self.run_iteration,
                    [
                        [
                            it,
                            start_time,
                            time_array,
                            time_unit,
                            save_dir,
                            (it in save_iterations and save_flag is True),
                            child_seeds[it - 1],
                            callback,
                        ]
                        for it in range(1, iterations + 1)
                    ],
                )

        if save_flag is True:
            # Merge monte carlo history variables from iterations
            save_dict = merge_monte_carlo_history(
                power_system=self.power_system,
                iteration_dicts=it_dicts,
            )
            # Save monte carlo history variables
            save_network_monte_carlo_history(
                power_system=self.power_system,
                save_dir=os.path.join(save_dir, "monte_carlo"),
                save_dict=save_dict,
            )



import numpy as np

from relsad.Time import Time
from relsad.utils import interpolate

from .Bus import Bus
from .Component import Component


class Production(Component):

    """
    Common class for production

    ...

    Attributes
    ----------
    name : string
        Name of the production unit
    bus : Bus
        The bus the production unit is connected to
    pprod_data : np.ndarray
        Array of active production data
    qprod_data : np.ndarray
        Array of reactive production data
    pprod : float
        The active power produced by the production unit [MW]
    qprod : float
        The reactive power produced by the production unit [MVar]
    pmax : float
        The maximum active power that can be produced by the production unit [MW]
    qmax : float
        The maximum reactive power that can be produced by the production unit [MVar]
    history : dict
        Dictonary attribute that stores the historic variables


    Methods
    ----------
    add_prod_data(pprod_data, qprod_data)
        Adds production data to the production component
    prepare_prod_data(time_indices)
        Prepares the production data for the current time step configuration
    add_prod(pprod, qprod)
        Adds production to the bus
    set_prod(curr_time)
        Decides how much active and reactive power that will be produced
    update_bus_prod()
        Updates the production on the bus with the amount of generated active and reactive power
        Sets the active production at the bus equal the active production at the bus minus the generated active power
        Sets the reactive production at the bus equal the reactive production at the bus minus the generated reactive power
    reset_prod()
        Resets the active and reactive production
    update_fail_status()
        Updates the fail status
    update_history(curr_time)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    print_status()
        Prints the status
    reset_status()
        Resets and sets the status of the class parameters

    """

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        bus: Bus,
        pmax: float = 10,
        qmax: float = 10,
    ):

        # Verify input
        if bus is None:
            raise Exception("Production unit must be connected to a Bus")
        if bus.parent_network is not None:
            raise Exception(
                "Production unit must be created before the bus is connected to a network"
            )
        if pmax < 0:
            raise Exception(
                "The maximum active power that can be produced must be positive"
            )
        if qmax < 0:
            raise Exception(
                "The maximum reactive power that can be produced must be positive"
            )

        self.name = name
        self.bus = bus
        bus.prod = self
        self.pprod_data = None
        self.qprod_data = None
        self.pprod = 0
        self.qprod = 0
        self.pmax = pmax
        self.qmax = qmax

        ## History
        self.history = {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Production(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, Production)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def add_prod_data(
        self,
        pprod_data: np.ndarray,
        qprod_data: np.ndarray = None,
    ):
        """
        Adds production data to the production component

        Parameters
        ----------
        pprod_data : np.ndarray
            Active power production array
        qprod_data : np.ndarray
            Reactive power production array

        Returns
        ----------
        None

        """
        self.pprod_data = pprod_data
        if qprod_data is None:
            self.qprod_data = np.zeros_like(pprod_data)
        else:
            self.qprod_data = qprod_data

    def prepare_prod_data(
        self,
        time_indices: np.ndarray,
    ):
        """
        Prepares the production data for the current time step configuration

        Parameters
        ----------
        time_indices : np.ndarray
            Time indices used to discretize the production data

        Returns
        ----------
        None

        """
        if self.pprod_data is None:
            raise Exception(
                "Active production data must be provided for {:s}".format(
                    self.name
                )
            )

        self.pprod_data = interpolate(
            array=self.pprod_data,
            time_indices=time_indices,
        )

        if self.qprod_data is None:
            raise Exception(
                "Reactive production data must be provided for {:s}".format(
                    self.name
                )
            )

        self.qprod_data = interpolate(
            array=self.qprod_data,
            time_indices=time_indices,
        )

    def add_prod(
        self,
        pprod: float,
        qprod: float,
    ):
        """
        Adds production to the bus

        Parameters
        ----------
        pprod : float
            Active power
        qprod : float
            Reactive power

        Returns
        ----------
        None

        """
        # MW and MVar
        self.pprod += pprod
        self.qprod += qprod

    def set_prod(self, inc_idx: int):
        """
        Decides how much active and reactive power that will be produced

        Parameters
        ----------
        inc_idx : int
            Index of current increment

        Returns
        ----------
        None

        """
        self.reset_prod()
        pprod = self.pprod_data[inc_idx]
        qprod = self.qprod_data[inc_idx]
        if pprod > self.pmax:
            pprod = self.pmax
        if qprod > self.qmax:
            qprod = self.qmax
        self.add_prod(pprod, qprod)
        self.update_bus_prod()

    def update_bus_prod(self):
        """
        Updates the production on the bus with the amount of generated active and reactive power
        Sets the active production at the bus equal the active production at the bus minus the generated active power
        Sets the reactive production at the bus equal the reactive production at the bus minus the generated reactive power

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.bus.pprod = self.pprod  # MW
        self.bus.qprod = self.qprod  # MVar
        self.bus.pprod_pu = self.pprod / self.bus.s_ref  # PU
        self.bus.qprod_pu = self.qprod / self.bus.s_ref  # PU

    def reset_prod(self):
        """
        Resets the active and reactive production

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.pprod = 0
        self.qprod = 0

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        curr_time : Time
            Current time

        Returns
        ----------
        None

        """

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """
        Prints the status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history = {}


if __name__ == "__main__":
    pass



import matplotlib.lines as mlines
import numpy as np

from relsad.StatDist import StatDist, StatDistType, UniformParameters
from relsad.Time import Time, TimeUnit
from relsad.utils import convert_yearly_fail_rate, random_choice

from .Bus import Bus
from .Component import Component


class Line(Component):
    r"""
    A class used to represent an electrical Line

    ...

    Attributes
    ----------
    name : str
        Name of the line
    is_backup : bool
        Decides if the line is a backup line
    fbus : Bus
        Sending bus
    tbus : Bus
        Receiving bus
    disconnectors : list
        List of disconnectors connected to the line
    circuitbreaker : CircuitBreaker
        Circuit breaker connected to the line
    parent_network : PowerNetwork
        The parent network of the line
    section : Section
        The section the line belongs to
    s_ref : float
        Reference apperet power [MVA]
    v_ref : float
        Reference voltage [kV]
    r_ref : float
        Reference resistance [Ohm]
    r : float
        Resistance [Ohm]
    x : float
        Reactance [Ohm]
    r_pu : float
        The pu value of the resistance
    x_pu : float
        The pu value of the reactance
    area : float
        The cross-sectional area [m^2]
    rho : float
        The resistivity of the line [Ohm*m]
    length : float
        Length of line [km]
    capacity : float
        The capacity of the line [MW]
    ploss : float
        The active power loss over the line [MW]
    qloss : float
        The reactive power loss over the line [MVar]
    fail_rate_per_year : float
        Failure rate per year [fault/year/km]
    repair_time_dist : StatDist
        The repair time of the line [hours/fault]
    connected : bool
        Indicates if the line is connected or disconnected
    failed : bool
        Failure status of the line
    remaining_outage_time : Time
        The remaining outage time of the line
    sensor : Sensor
        The sensor(s) connected to the line
    history : dict
        Dictonary attribute that stores the historic variables


    Methods
    -------
    set_backup()
        Sets the backup lines and opens the disconnectors connected to the backup line
    disconnect()
        Disconnects a line and removes the line for the list of lines
    connect()
        Connects a line and append the line to the list of lines
    draw_repair_time(dt)
        Decides and returns the repair time of the line based on a statistical distribution
    fail(dt)
        Sets the fail status of the line to True and opens the connected disconnectors and the connected circuit breaker
    not_fail()
        Sets the fail status of the line to False and closes the connected disconnectors and connected circuit breaker
    change_direction()
        Changes the direction of the line
    update_fail_status(dt)
        Updates the fail status of the line
    get_line_load()
        Returns the flow over the line in PU values
    get_line_loading()
        Returns the loading of the line in percentage
    print_status()
        Prints the line status
    get_disconnetors()
        Returns the connected disconnectors
    add_parent_network(network)
        Adds the parent network to the line
    initialize_history()
        Initialize the history variables
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    reset_status()
        Resets and sets the status of the class parameters
    reset_load_flow_data()
        Resets the variables used in the load flow analysis
    get_switches()
        Returns the switches on the line
    """
    lineCount = 0

    ## Visual attributes
    linestyle = "-"
    handle = mlines.Line2D([], [], linestyle=linestyle)

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        fbus: Bus,
        tbus: Bus,
        r: float,  # Ohm
        x: float,  # Ohm
        repair_time_dist: StatDist = StatDist(
            stat_dist_type=StatDistType.UNIFORM_FLOAT,
            parameters=UniformParameters(
                min_val=0.0,
                max_val=0.0,
            ),
        ),
        s_ref: float = 1,  # MVA
        v_ref: float = 12.66,  # kV
        rho: float = 1.72e-8,  # resistivity [Ohm*m]
        area: float = 64.52e-6,  # cross-sectional area [m**2]
        fail_rate_density_per_year: float = 0.0,  # fails/(km*year)
        capacity: float = 100,  # MW
        connected=True,
    ):

        # Verify input
        if fbus is None:
            raise Exception("Line must have a origin bus")
        if tbus is None:
            raise Exception("Line must have a target bus")
        if r < 0:
            raise Exception("r must be positive")
        if x < 0:
            raise Exception("x must be positive")
        if repair_time_dist is None:
            raise Exception("The line must have a repair time distribution")
        if s_ref < 0:
            raise Exception("s_ref must be positive")
        if v_ref < 0:
            raise Exception("v_ref must be positive")
        if rho < 0:
            raise Exception("The resistivity must be positive")
        if area < 0:
            raise Exception("The line section area must be positive")
        if fail_rate_density_per_year < 0:
            raise Exception(
                "The failure rate density per year must be positive"
            )
        if capacity < 0:
            raise Exception("The line capacity must be positive")

        ## Informative attributes
        self.name = name

        ## Backup
        self.is_backup = False

        ## Topological attributes
        self.fbus = fbus
        self.tbus = tbus
        fbus.connected_lines.append(self)
        tbus.connected_lines.append(self)
        tbus.toline = self
        tbus.toline_list.append(self)
        fbus.fromline = self
        fbus.fromline_list.append(self)
        fbus.nextbus.append(self.tbus)
        self.disconnectors = list()
        self.circuitbreaker = None
        self.parent_network = None
        self.section = None
        Line.lineCount += 1

        ##  Power flow attributes
        self.s_ref = s_ref
        self.v_ref = v_ref
        self.r_ref = v_ref**2 / s_ref
        self.r = r
        self.x = x
        self.r_pu = r / self.r_ref
        self.x_pu = x / self.r_ref
        self.area = area
        self.rho = rho
        self.length = r * area / rho * 1e-3  # km
        self.capacity = capacity  # MW
        self.ploss = 0.0
        self.qloss = 0.0

        ## Reliabilility attributes
        self.fail_rate_per_year = (
            fail_rate_density_per_year * self.length
        )  # failures per year
        self.repair_time_dist = repair_time_dist

        ## Status attribute
        self.connected = connected
        self.failed = False
        self.remaining_outage_time = Time(0)

        ## Communication

        self.sensor = None

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Line(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, Line)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def set_backup(self):
        """
        Sets the backup lines and opens the disconnectors connected to the backup line

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_backup = True
        for discon in self.disconnectors:
            discon.open()

    def disconnect(self):
        """
        Disconnects a line and removes the line from the list of lines

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        if self.connected:
            self.connected = False
            self.linestyle = "--"
            self.fbus.fromline_list.remove(self)
            if self.fbus.fromline == self:
                if len(self.fbus.fromline_list) > 0:
                    self.fbus.fromline = next(iter(self.fbus.fromline_list))
                else:
                    self.fbus.fromline = None
            self.tbus.toline_list.remove(self)
            if self.tbus.toline == self:
                if len(self.tbus.toline_list) > 0:
                    self.tbus.toline = next(iter(self.tbus.toline_list))
                else:
                    self.tbus.toline = None
            self.fbus.nextbus.remove(self.tbus)

    def connect(self):
        """
        Connects a line and append the line to the list of lines

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        if not self.connected:
            self.connected = True
            self.linestyle = "-"
            self.tbus.toline = self
            self.tbus.toline_list.append(self)
            self.fbus.fromline = self
            self.fbus.fromline_list.append(self)
            self.fbus.nextbus.append(self.tbus)

    def draw_repair_time(self, dt: Time):
        """
        Decides and returns the repair time of the line based on a statistical distribution

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        return Time(
            self.repair_time_dist.draw(
                random_instance=self.ps_random,
                size=1,
            )[0],
            dt.unit,
        )

    def fail(self, dt: Time):
        """
        Sets the fail status of the line to False and opens the connected disconnectors and the connected circuit breaker

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.failed = True
        self.parent_network.failed_line = True
        self.remaining_outage_time = self.draw_repair_time(dt)
        if self.connected:
            # Relay

            # Disconnects parent network from its parent network
            self.parent_network.connected_line.circuitbreaker.open()

            # All child networks are disconnected from the parent network
            if hasattr(self.parent_network, "child_network_list"):
                if self.parent_network.child_network_list is not None:
                    for (
                        child_network
                    ) in self.parent_network.child_network_list:
                        child_network.connected_line.circuitbreaker.open()

    def not_fail(self):
        """
        Sets the fail status of the line to False and closes the connected disconnectors and connected circuit breaker

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        if (
            sum([line.failed for line in self.parent_network.lines]) == 1
            and self.failed
        ):
            self.parent_network.failed_line = False
        self.failed = False

    def change_direction(self):
        """
        Changes the direction of the line

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.fbus.fromline_list.remove(self)
        self.tbus.fromline_list.append(self)
        self.tbus.toline_list.remove(self)
        self.fbus.toline_list.append(self)
        if self.fbus.fromline == self:
            self.fbus.fromline = (
                next(iter(self.fbus.fromline_list))
                if len(self.fbus.fromline_list) > 0
                else None
            )
        if self.tbus.toline == self:
            self.tbus.toline = (
                next(iter(self.tbus.toline_list))
                if len(self.tbus.toline_list) > 0
                else None
            )
        self.fbus.toline = self
        self.tbus.fromline = self
        self.fbus.nextbus.remove(self.tbus)
        self.tbus.nextbus.append(self.fbus)
        bus = self.fbus
        self.fbus = self.tbus
        self.tbus = bus

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status of the line

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.is_backup:
            for discon in self.disconnectors:
                if not discon.is_open:
                    discon.open()
        if self.failed:
            self.remaining_outage_time -= dt
            if self.remaining_outage_time <= Time(0):
                self.not_fail()
                self.parent_network.controller.check_components = True
                self.remaining_outage_time = Time(0)
        else:
            p_fail = convert_yearly_fail_rate(self.fail_rate_per_year, dt)
            if random_choice(self.ps_random, p_fail):
                self.fail(dt)
            else:
                self.not_fail()

    def get_line_load(self):
        """
        Returns the flow over the line in PU values

        Parameters
        ----------
        None

        Returns
        ----------
        p_from : float
            The active power sent from the line [MW]
        q_from : float
            The reactive power sent from the line [MVar]
        p_to : float
            The active power sent to the line [MW]
        q_to : float
            The reactive power sent to the line [MVar]

        """

        def uij(gij, bij, tetai, tetaj):
            return gij * np.sin(tetai - tetaj) - bij * np.cos(tetai - tetaj)

        def tij(gij, bij, tetai, tetaj):
            return gij * np.cos(tetai - tetaj) + bij * np.sin(tetai - tetaj)

        def bij(R, X):
            return (1.0 / complex(R, X)).imag

        def gij(R, X):
            return (1.0 / complex(R, X)).real

        if self.connected:
            fbus = self.fbus
            tbus = self.tbus
            bsh = 0.0  # No shunts included so far
            teta1 = fbus.voang
            teta2 = tbus.voang
            v1 = fbus.vomag
            v2 = tbus.vomag
            b = bij(self.r_pu, self.x_pu)
            g = gij(self.r_pu, self.x_pu)

            p_from = g * v1 * v1 - v1 * v2 * tij(g, b, teta1, teta2)
            p_to = g * v2 * v2 - v1 * v2 * tij(g, b, teta2, teta1)
            q_from = -(b + bsh) * v1 * v1 - v1 * v2 * uij(g, b, teta1, teta2)
            q_to = -(b + bsh) * v2 * v2 - v1 * v2 * uij(g, b, teta2, teta1)

            return p_from, q_from, p_to, q_to
        else:
            return 0, 0, 0, 0

    def get_line_loading(self):
        """
        Returns the loading of the line in percentage

        Parameters
        ----------
        None

        Returns
        ----------
        line_loading : float
            The line loading

        """

        p_from = self.get_line_load()[0]
        line_loading = abs(p_from) / (self.capacity / self.s_ref) * 100
        return line_loading

    def print_status(self):
        """
        Prints the line status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        print(
            "name: {:5s}, failed={}, connected={}".format(
                self.name, self.failed, self.connected
            )
        )

    def get_disconnectors(self):
        """
        Returns the connected disconnectors

        Parameters
        ----------
        None

        Returns
        ----------
        disconnectors : Disconnector
            Disconnector connected to the line

        """
        return self.disconnectors

    def add_parent_network(self, network):
        """
        Adds the parent network to the line

        Parameters
        ----------
        network : PowerNetwork
            The parent network of the line

        Returns
        ----------
        None

        """

        self.parent_network = network

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["p_from"] = {}
        self.history["q_from"] = {}
        self.history["p_to"] = {}
        self.history["q_to"] = {}
        self.history["remaining_outage_time"] = {}
        self.history["failed"] = {}
        self.history["line_loading"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            The vurrent time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            p_from, q_from, p_to, q_to = self.get_line_load()
            self.history["p_from"][time] = p_from * self.s_ref
            self.history["q_from"][time] = q_from * self.s_ref
            self.history["p_to"][time] = p_to * self.s_ref
            self.history["q_to"][time] = q_to * self.s_ref
            self.history["remaining_outage_time"][
                time
            ] = self.remaining_outage_time.get_unit_quantity(curr_time.unit)
            self.history["failed"][time] = self.failed
            self.history["line_loading"][time] = self.get_line_loading()

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.remaining_outage_time = Time(0)

        self.not_fail()
        if save_flag:
            self.initialize_history()

    def reset_load_flow_data(self):
        """
        Resets the variables used in the load flow analysis

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.ploss = 0
        self.qloss = 0

    def get_switches(self):
        """
        Returns the switches on the line

        Parameters
        ----------
        None

        Returns
        ----------
        switches
            The switches on the line

        """
        switches = (
            self.disconnectors + [self.circuitbreaker]
            if self.circuitbreaker is not None
            else self.disconnectors
        )
        return switches


if __name__ == "__main__":
    pass



from relsad.examples.tutorial.system_components import *

from relsad.network.components import (
    MainController,
    Sensor,
    IntelligentSwitch,
)

C1 = MainController(name="C1")

Isw1 = IntelligentSwitch(
    name="Isw1",
    disconnector=DL2a,
)

S1 = Sensor(
    name="S1",
    line=L2,
)



from relsad.examples.tutorial.system_components import *

from relsad.network.systems import ICTNetwork
from relsad.network.components import (
    ICTNode,
    ICTLine,
)

from relsad.network.systems import PowerSystem

ps = PowerSystem(controller=C1)

# ICT nodes
ICTNC1 = ICTNode(
    name="ICTNC1",
)
ICTNISW1 = ICTNode(
    name="ICTNISW1",
)
ICTNS1 = ICTNode(
    name="ICTNS1",
)

# ICT lines
ICTL1 = ICTLine(
    name="ICTL1",
    fnode=ICTNC1,
    tnode=ICTNISW1,
)
ICTL2 = ICTLine(
    name="ICTL2",
    fnode=ICTNC1,
    tnode=ICTNS1,
)
ICTL3 = ICTLine(
    name="ICTL3",
    fnode=ICTNS1,
    tnode=ICTNISW1,
)

# ICT network
ict_network = ICTNetwork(ps)
ict_network.add_nodes(
    [
        ICTNC1,
        ICTNISW1,
        ICTNS1,
    ]
)
ict_network.add_lines(
    [
        ICTL1,
        ICTL2,
        ICTL3,
    ]
)



from .array import interpolate

from .compare import eq

from .definitions import INF

from .random import (
    get_random_instance,
    random_choice,
    convert_yearly_fail_rate,
)

from .set import (
    unique,
    subtract,
    intersection,
)

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



from relsad.network.components import (
    Bus,
    Line,
    Disconnector,
    CircuitBreaker,
    ManualMainController,
)
from relsad.Time import (
    Time,
    TimeUnit,
    TimeStamp,
)
from relsad.StatDist import (
    StatDist,
    StatDistType,
    NormalParameters,
)


# Failure rate and outage time of the transformer on the bus
# are not necessary to add, this can be added on each bus.
# Their default values are 0 and Time(0) respectively.

B1 = Bus(
    name="B1",
    n_customers=0,
    coordinate=[0, 0],
)
B2 = Bus(
    name="B2",
    n_customers=1,
    coordinate=[1, 0],
)

B3 = Bus(
    name="B3",
    n_customers=1,
    coordinate=[2, 1],
)

B4 = Bus(
    name="B4",
    n_customers=1,
    coordinate=[2, 0],
)

B5 = Bus(
    name="B5",
    n_customers=1,
    coordinate=[3, 0],
)

B6 = Bus(
    name="B6",
    n_customers=1,
    coordinate=[3, 1],
)

# Failure rate and outage time of the lines can be added to each line.
# The default value of the line failure rate is 0, while the default
# outage time is 0 (Uniform float distribution with max/min values of 0).

# For adding statistical distributions, in this case a
# truncated normal distribution:

line_stat_repair_time_dist = StatDist(
    stat_dist_type=StatDistType.TRUNCNORMAL,
    parameters=NormalParameters(
        loc=1.25,
        scale=1,
        min_val=0.5,
        max_val=2,
    ),
)

fail_rate_line = 0.07

L1 = Line(
    name="L1",
    fbus=B1,
    tbus=B2,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)
L2 = Line(
    name="L2",
    fbus=B2,
    tbus=B3,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)
L3 = Line(
    name="L3",
    fbus=B2,
    tbus=B4,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)
L4 = Line(
    name="L4",
    fbus=B4,
    tbus=B5,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)
L5 = Line(
    name="L5",
    fbus=B3,
    tbus=B6,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)

# Backup line

L6 = Line(
    name="L6",
    fbus=B4,
    tbus=B6,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)

# Set L6 as a backup line

L6.set_backup()

E1 = CircuitBreaker(
    name="E1",
    line=L1,
)

DL1a = Disconnector(
    name="L1a",
    line=L1,
    bus=B1,
)
DL1b = Disconnector(
    name="L1b",
    line=L1,
    bus=B2,
)
DL2a = Disconnector(
    name="L2a",
    line=L2,
    bus=B2,
)
DL2b = Disconnector(
    name="L2b",
    line=L2,
    bus=B3,
)
DL3a = Disconnector(
    name="L3a",
    line=L3,
    bus=B2,
)
DL3b = Disconnector(
    name="L3b",
    line=L3,
    bus=B4,
)
DL4a = Disconnector(
    name="L4a",
    line=L4,
    bus=B4,
)
DL4b = Disconnector(
    name="L4b",
    line=L4,
    bus=B5,
)
DL5a = Disconnector(
    name="L5a",
    line=L5,
    bus=B3,
)
DL5b = Disconnector(
    name="L5b",
    line=L5,
    bus=B6,
)

# For backup line
DL6a = Disconnector(
    name="L6a",
    line=L6,
    bus=B4,
)
DL6b = Disconnector(
    name="L6b",
    line=L6,
    bus=B6,
)

C1 = ManualMainController(name="C1", sectioning_time=Time(0))



import matplotlib.lines as mlines
import numpy as np

from relsad.Time import Time, TimeUnit
from relsad.utils import convert_yearly_fail_rate, random_choice

from .Component import Component
from .Controller import Controller


class ManualMainController(Component):

    """
    Common class for manual main controller

    ...

    Attributes
    ----------
    name : string
        Name of the manual main controller
    sectioning_time : Time
        The average sectioning time of the system
    power_system : PowerSystem
        The associated power system
    distribution_controllers : list
        List of controllers from the distribution systems in the power system
    microgrid_controllers : list
        List of controllers from the microgrids in the power system


    Methods
    ----------
    add_distribution_controller(controller)
        Adds distribution controllers from connected distribution systems to a list and sets the manual sectioning time for the controller
    add_microgrid_controller(controller)
        Adds microgird controllers from connected microgrids to a list and sets the manual sectioning time for the controller
    run_control_loop(curr_time, dt)
        Runs the manual control loop for the distribution system controllers and the microgird controllers
    update_fail_status(dt)
        Updates the failure status
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    print_status()
        Prints the status
    reset_status(save_flag)
        Resets and sets the status of the class parameters
    initialize_history()
        Initializes the history variables
    """

    def __init__(
        self,
        name: str,
        sectioning_time: Time,
    ):

        self.name = name
        self.sectioning_time = sectioning_time
        self.power_system = None

        self.distribution_controllers = list()
        self.microgrid_controllers = list()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"ManualMainController(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(
                other, ManualMainController
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def add_distribution_controller(self, controller):
        """
        Adds distribution controllers from connected distribution systems to a list
        and sets the manual sectioning time for the controller

        Parameters
        ----------
        controller : Controller
            Distribution system controller

        Returns
        ----------
        None

        """
        self.distribution_controllers.append(controller)
        controller.manual_sectioning_time = self.sectioning_time

    def add_microgrid_controller(self, controller):
        """
        Adds microgird controllers from connected microgrids to a list and sets the manual sectioning time for the controller

        Parameters
        ----------
        controller : Controller
            Microgrid controller

        Returns
        ----------
        None

        """
        self.microgrid_controllers.append(controller)
        controller.manual_sectioning_time = self.sectioning_time

    def run_control_loop(self, curr_time: Time, dt: Time):
        """
        Runs the manual control loop for the distribution system controllers and the microgird controllers

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        for controller in self.distribution_controllers:
            controller.run_manual_control_loop(curr_time, dt)
        for controller in self.microgrid_controllers:
            controller.run_manual_control_loop(curr_time, dt)

    def update_fail_status(self, dt: Time):
        """
        Updates the failure status

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        pass

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        pass

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            Manual controller attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        pass

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        pass

    def print_status(self):
        """
        Prints the status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        pass

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        pass

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        pass



import matplotlib.lines as mlines
import numpy as np

from relsad.network.containers import SectionState
from relsad.Time import Time, TimeUnit
from relsad.topology.ICT.dfs import is_connected
from relsad.utils import convert_yearly_fail_rate, random_choice, unique

from .Component import Component
from .Controller import Controller, ControllerState
from .ICTNode import ICTNode


class DistributionController(Component, Controller):

    """
    Common base class for distribution controllers

    ...

    Attributes
    ----------
    name : string
        Name of the controller
    ict_node : ICTNode
        The ICT node connected to the controller
    fail_rate : float
        The failure rate of the distribution system controller
    state : ControllerState
        The state of the distribution system controller
    sectioning_time : Time
        The sectioning time of the system
    check_components : bool
        Flag used to communicate whether the fail status of the system
        components should be checked during a control loop
    manual_sectioning_time : Time
        Sectioning time when performed manually
    parent_controller : Controller
        The parent controller of the distribution system controller
    power_network : PowerNetwork
        The power network the controller is connected to
    ict_network : ICTNetwork
        The ICT network the controller belongs to
    sensors : list
        List of sensors included in the distribution system
    failed_sections : list
        List of failed sections in the distribution system
    history : dict
        Dictonary attribute that stores the historic variables
    monte_carlo_history : dict
        Dictonary containing the history variables from the Monte Carlo simulation

    Methods
    ----------
    check_circuitbreaker(curr_time, dt)
        Checks if the circuitbreaker in the distribution system is open.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.
    check_circuitbreaker_manually(curr_time, dt)
        Checks if the circuitbreaker in the distribution system is open manually.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.
    disconnect_failed_sections()
        Disconnects the failed sections in the distribution system
    check_sensors(curr_time, dt)
        Loops through the sections connected to the controller determining which
        sensors have failed. Performs actions according to the sensor status
        in the respective section.
        If a section was disconnected and no longer includes any failed sensor,
        it is connected.
        If a section was connected and now includes a failed sensor, it is disconnected.
        The total sectioning time is summed from each section.
    run_control_loop(curr_time, dt)
        System control check, determines if components have failed and performes
        the required action
    check_lines_manually(curr_time)
        Loops through the sections connected to the controller determining
        which lines have failed manually. Performs actions according
        to the line status in the respective section.
        If a section was disconnected and no longer includes any failed lines,
        it is connected.
        If a section was connected and now includes a failed line, it is disconnected.
        The total sectioning time is summed from each section.
    run_manual_control_loop(curr_time, dt)
        Manual system control check, determines if components have failed and
        performes the required action.
    set_sectioning_time(sectioning_time)
        Sets the sectiom time of the distribuiton system based on the max value
        of the distribution sectioning time and the sectioning time set by the
        controller
    spread_sectioning_time_to_children()
        Gives the children power networks of the distribution system the same
        sectioning time
    update_fail_status(dt)
        Updated the fail status of the controller
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    print_status()
        Prints the status
    reset_status(save_flag)
        Resets and sets the status of the class parameters
    initialize_history()
        Initializes the history variables
    """

    def __init__(
        self,
        name: str,
        power_network,
        fail_rate: float = 0,
        state: ControllerState = ControllerState.OK,
    ):

        self.name = name
        self.ict_node = None
        self.fail_rate = fail_rate
        self.state = state
        self.sectioning_time = Time(0)
        self.check_components = False

        self.manual_sectioning_time = None

        self.parent_controller = None

        self.power_network = power_network
        self.ict_network = None

        self.sensors = []

        self.failed_sections = []

        ## History
        self.history = {}
        self.monte_carlo_history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"DistributionController(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(
                other, DistributionController
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def check_circuitbreaker(self, curr_time: Time, dt: Time):
        """
        Checks if the circuitbreaker in the distribution system is open.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.power_network.connected_line.circuitbreaker.is_open:
            if self.sectioning_time <= Time(0):
                self.disconnect_failed_sections()
                if not self.power_network.connected_line.failed:
                    # Sectioning time finished
                    self.power_network.connected_line.circuitbreaker.close()
                    self.power_network.connected_line.section.connect(dt, self)
                    self.failed_sections = []

    def check_circuitbreaker_manually(self, curr_time: Time, dt: Time):
        """
        Checks if the circuitbreaker in the distribution system is open manually.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.power_network.connected_line.circuitbreaker.is_open:
            if self.sectioning_time <= Time(0):
                self.disconnect_failed_sections()
                if (
                    not self.power_network.connected_line.failed
                    and self.power_network.connected_line
                    not in [
                        line
                        for section in self.failed_sections
                        for line in section.lines
                    ]
                ):
                    # Sectioning time finished
                    self.power_network.connected_line.circuitbreaker.close()
                    self.power_network.connected_line.section.connect_manually()
                    self.failed_sections = []

    def disconnect_failed_sections(self):
        """
        Disconnects the failed sections in the distribution system

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for section in self.failed_sections:
            section.disconnect()

    def check_sensors(self, curr_time: Time, dt: Time):
        """
        Loops through the sections connected to the controller determining
        which sensors have failed. Performs actions according
        to the sensor status in the respective section.

        If a section was disconnected and no longer includes any failed
        sensor, it is connected.

        If a section was connected and now includes a failed sensor,
        it is disconnected.

        The total sectioning time is summed from each section.


        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        connected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.CONNECTED
        ]
        disconnected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.DISCONNECTED
        ]
        # Loop disconnected sections
        for section in disconnected_sections:
            sensors = unique(
                [
                    switch.line.sensor
                    for switch in section.switches
                    if switch.line.sensor is not None
                ]
            )
            num_fails = 0
            need_manual_attention = False
            # Loop sensors and count failed ones
            for sensor in sensors:
                # If no ICT network
                if self.ict_node is None:
                    (
                        repair_time,
                        line_fail_status,
                    ) = sensor.get_line_fail_status(dt)
                    self.sectioning_time += repair_time
                # If both components have ICT nodes
                elif self.ict_node is not None and sensor.ict_node is not None:
                    # If the ICT nodes are connected to each other
                    if is_connected(
                        node_1=self.ict_node,
                        node_2=sensor.ict_node,
                        network=self.ict_network,
                    ):
                        (
                            repair_time,
                            line_fail_status,
                        ) = sensor.get_line_fail_status(dt)
                        self.sectioning_time += repair_time
                    # If the ICT nodes are not connected to each other
                    else:
                        need_manual_attention = True
                        line_fail_status = sensor.line.failed
                # If no ICT node on sensor
                else:
                    need_manual_attention = True
                    line_fail_status = sensor.line.failed
                num_fails += 1 if line_fail_status is True else 0
            if need_manual_attention is True:
                self.sectioning_time += self.manual_sectioning_time
            if num_fails == 0:
                section.connect(dt, self)
                if section in self.failed_sections:
                    self.failed_sections.remove(section)
        # Loop connected sections
        for section in connected_sections:
            sensors = unique(
                [
                    switch.line.sensor
                    for switch in section.switches
                    if switch.line.sensor is not None
                ]
            )
            num_fails = 0
            need_manual_attention = False
            # Loop sensors and count failed ones
            for sensor in sensors:
                # If no ICT network
                if self.ict_node is None:
                    (
                        repair_time,
                        line_fail_status,
                    ) = sensor.get_line_fail_status(dt)
                    self.sectioning_time += repair_time
                # If both components have ICT nodes
                elif self.ict_node is not None and sensor.ict_node is not None:
                    # If the ICT nodes are connected to each other
                    if is_connected(
                        node_1=self.ict_node,
                        node_2=sensor.ict_node,
                        network=self.ict_network,
                    ):
                        (
                            repair_time,
                            line_fail_status,
                        ) = sensor.get_line_fail_status(dt)
                        self.sectioning_time += repair_time
                    # If the ICT nodes are not connected to each other
                    else:
                        need_manual_attention = True
                        line_fail_status = sensor.line.failed
                # If no ICT node on sensor
                else:
                    need_manual_attention = True
                    line_fail_status = sensor.line.failed
                num_fails += 1 if line_fail_status else 0
            if need_manual_attention is True:
                self.sectioning_time += self.manual_sectioning_time
            if num_fails > 0:
                section.state = SectionState.DISCONNECTED
                self.sectioning_time += section.get_disconnect_time(dt, self)
                self.failed_sections.append(section)
                self.failed_sections = unique(self.failed_sections)

    def run_control_loop(self, curr_time: Time, dt: Time):
        """
        System control check, determines if components have failed and
        performes the required action.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        # Update current sectioning time
        self.sectioning_time = (
            self.sectioning_time - dt
            if self.sectioning_time > Time(0)
            else Time(0)
        )
        if (
            self.power_network.connected_line.circuitbreaker.is_open
            and self.sectioning_time <= Time(0)
        ):
            # Failure occured in current time step
            self.check_components = True
        if self.check_components:
            # Check sensor status
            self.check_sensors(curr_time, dt)
            self.spread_sectioning_time_to_children()
            self.check_components = False
        # Check circuitbreaker status
        self.check_circuitbreaker(curr_time, dt)

    def check_lines_manually(self, curr_time):
        """
        Loops manually through the sections connected to the controller determining
        which lines have failed. Performs actions according
        to the line status in the respective section.

        If a section was disconnected and no longer includes any failed
        lines, it is connected.

        If a section was connected and now includes a failed line,
        it is disconnected.

        The total sectioning time is summed from each section.

        Parameters
        ----------
        curr_time : Time
            The current time

        Returns
        ----------
        None

        """
        connected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.CONNECTED
        ]
        disconnected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.DISCONNECTED
        ]
        # Loop disconnected sections
        for section in disconnected_sections:
            if sum(x.failed for x in section.lines) == 0:
                section.connect_manually()
                if section in self.failed_sections:
                    self.failed_sections.remove(section)
        # Loop connected sections
        for section in connected_sections:
            if sum(x.failed for x in section.lines) > 0:
                section.state = SectionState.DISCONNECTED
                self.failed_sections.append(section)
                self.failed_sections = unique(self.failed_sections)
                self.sectioning_time = self.manual_sectioning_time
                for line in section.lines:
                    line.remaining_outage_time += self.sectioning_time

    def run_manual_control_loop(self, curr_time: Time, dt: Time):
        """
        Manual system control check, determines if components have failed and
        performes the required action.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        # Update current sectioning time
        self.sectioning_time = (
            self.sectioning_time - dt
            if self.sectioning_time > Time(0)
            else Time(0)
        )
        if (
            self.power_network.connected_line.circuitbreaker.is_open
            and self.sectioning_time <= Time(0)
        ):
            # Failure occured in current time step
            self.check_components = True
        if self.check_components:
            # Check line status
            self.check_lines_manually(curr_time)
            self.spread_sectioning_time_to_children()
            self.check_components = False
        # Check circuitbreaker status
        self.check_circuitbreaker_manually(curr_time, dt)

    def set_sectioning_time(self, sectioning_time):
        """
        Sets the sectioning time of the distribuiton system based on
        the max value of the distribution sectioning time and the
        sectioning time set by the controller

        Parameters
        ----------
        sectioning_time : Time
            The sectioning time of the distribution system

        Returns
        ----------
        None

        """
        self.sectioning_time = max(
            self.sectioning_time,
            sectioning_time,
        )

    def spread_sectioning_time_to_children(self):
        """
        Returns the children power network of the distribution system
        the same sectioning time

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for child_network in self.power_network.child_network_list:
            if (
                child_network.controller.power_network.connected_line.circuitbreaker.is_open
            ):
                child_network.controller.set_parent_sectioning_time(
                    self.sectioning_time
                )

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status of the controller

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        pass

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["sectioning_time"][
                time
            ] = self.sectioning_time.get_unit_quantity(curr_time.unit)

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            Controller attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        pass

    def print_status(self):
        """
        Prints the status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.sectioning_time = Time(0)
        if save_flag:
            self.initialize_history()

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["sectioning_time"] = {}
